@@ -116,12 +116,17 @@ typedef enum RateUnits {
  * and contains only C elements so it can be manipulated
  * by either C or C++.
  */
+#define PEERBUFSIZE 80
 typedef struct thread_Settings {
     // Pointers
     char*  mFileName;               // -F
     char*  mHost;                   // -c
     char*  mLocalhost;              // -B
     char*  mOutputFileName;         // -o
+    char*  mIfrname;                 // %<device> name
+    char*  mSSMMulticastStr;        // --ssm-host
+    char*  mIsochronousStr;         // --isochronous
+    char*  mUDPHistogramStr;        // --udp-histogram
     FILE*  Extractor_file;
     ReportHeader*  reporthdr;
     MultiHeader*   multihdr;
@@ -134,6 +139,12 @@ typedef struct thread_Settings {
     SOCKET mSock;
 #else
     int mSock;
+#endif
+#if defined(HAVE_LINUX_FILTER_H) && defined(HAVE_AF_PACKET)
+    int mSockDrop;
+#endif
+#ifdef HAVE_UDPTRIGGERS
+    int mSockIoctl;
 #endif
     int Extractor_size;
     int mBufLen;                    // -l
@@ -187,11 +198,29 @@ typedef struct thread_Settings {
     iperf_sockaddr local;
     Socklen_t size_local;
     nthread_t mTID;
+    int incrdstip;
     char* mCongestion;
-    char peerversion[80];
+    char peerversion[PEERBUFSIZE];
+    int mUDPbins;
+    int mUDPbinsize;
+    unsigned short mUDPunits;
+    unsigned short mUDPci_lower;
+    unsigned short mUDPci_upper;
 #if defined( HAVE_WIN32_THREAD )
     HANDLE mHandle;
 #endif
+#ifdef HAVE_ISOCHRONOUS
+    double mFPS; //frames per second
+    double mMean; //variable bit rate mean
+    int mJitterBufSize; //Server jitter buffer size, units is frames
+    double mBurstIPG; //Interpacket gap
+#endif
+    int l4offset; // used in l2 mode to offset the raw packet
+    int l4payloadoffset;
+    int recvflags; // used to set recv flags,e.g. MSG_TRUNC with L
+    struct timeval thread_synctime;
+    double mTxSyncInterval;
+    double mVariance; //vbr variance
 } thread_Settings;
 
 /*
@@ -234,12 +263,21 @@ typedef struct thread_Settings {
 #define FLAG_BWSET          0x01000000
 #define FLAG_ENHANCEDREPORT 0x02000000
 #define FLAG_SERVERMODETIME 0x04000000
+#define FLAG_SSM_MULTICAST  0x08000000
 /*
  * Extended flags
  */
 #define FLAG_PEERVER        0x00000001
 #define FLAG_SEQNO64        0x00000002
 #define FLAG_REVERSE        0x00000004
+#define FLAG_ISOCHRONOUS    0x00000008
+#define FLAG_UDPTRIGGERS    0x00000010
+#define FLAG_UDPHISTOGRAM   0x00000020
+#define FLAG_L2LENGTHCHECK  0x00000100
+#define FLAG_TXSYNC         0x00000200
+#define FLAG_INCRDSTIP      0x00000400
+#define FLAG_VARYLOAD       0x00000800
+
 
 #define isBuflenSet(settings)      ((settings->flags & FLAG_BUFLENSET) != 0)
 #define isCompat(settings)         ((settings->flags & FLAG_COMPAT) != 0)
@@ -256,6 +294,7 @@ typedef struct thread_Settings {
 #define isModeTime(settings)       ((settings->flags & FLAG_MODETIME) != 0)
 #define isReport(settings)         ((settings->flags & FLAG_REPORTSETTINGS) != 0)
 #define isMulticast(settings)      ((settings->flags & FLAG_MULTICAST) != 0)
+#define isSSMMulticast(settings)   ((settings->flags & FLAG_SSM_MULTICAST) != 0)
 // Active Low for Reports
 #define isSettingsReport(settings) ((settings->flags & FLAG_NOSETTREPORT) == 0)
 #define isConnectionReport(settings)  ((settings->flags & FLAG_NOCONNREPORT) == 0)
@@ -270,9 +309,16 @@ typedef struct thread_Settings {
 #define isBWSet(settings)          ((settings->flags & FLAG_BWSET) != 0)
 #define isEnhanced(settings)       ((settings->flags & FLAG_ENHANCEDREPORT) != 0)
 #define isServerModeTime(settings) ((settings->flags & FLAG_SERVERMODETIME) != 0)
-#define isPeerVerDetect(settings)       ((settings->flags_extend & FLAG_PEERVER) != 0)
+#define isPeerVerDetect(settings)  ((settings->flags_extend & FLAG_PEERVER) != 0)
 #define isSeqNo64b(settings)       ((settings->flags_extend & FLAG_SEQNO64) != 0)
-#define isReverse(settings)       ((settings->flags_extend & FLAG_REVERSE) != 0)
+#define isReverse(settings)        ((settings->flags_extend & FLAG_REVERSE) != 0)
+#define isIsochronous(settings)    ((settings->flags_extend & FLAG_ISOCHRONOUS) != 0)
+#define isUDPTriggers(settings)    ((settings->flags_extend & FLAG_UDPTRIGGERS) != 0)
+#define isUDPHistogram(settings)   ((settings->flags_extend & FLAG_UDPHISTOGRAM) != 0)
+#define isL2LengthCheck(settings)  ((settings->flags_extend & FLAG_L2LENGTHCHECK) != 0)
+#define isIncrDstIP(settings)       ((settings->flags_extend & FLAG_INCRDSTIP) != 0)
+#define isTxSync(settings)         ((settings->flags_extend & FLAG_TXSYNC) != 0)
+#define isVaryLoad(settings)         ((settings->flags_extend & FLAG_VARYLOAD) != 0)
 
 #define setBuflenSet(settings)     settings->flags |= FLAG_BUFLENSET
 #define setCompat(settings)        settings->flags |= FLAG_COMPAT
@@ -289,6 +335,7 @@ typedef struct thread_Settings {
 #define setModeTime(settings)      settings->flags |= FLAG_MODETIME
 #define setReport(settings)        settings->flags |= FLAG_REPORTSETTINGS
 #define setMulticast(settings)     settings->flags |= FLAG_MULTICAST
+#define setSSMMulticast(settings)  settings->flags |= FLAG_SSM_MULTICAST
 #define setNoSettReport(settings)  settings->flags |= FLAG_NOSETTREPORT
 #define setNoConnReport(settings)  settings->flags |= FLAG_NOCONNREPORT
 #define setNoDataReport(settings)  settings->flags |= FLAG_NODATAREPORT
@@ -301,9 +348,16 @@ typedef struct thread_Settings {
 #define setBWSet(settings)         settings->flags |= FLAG_BWSET
 #define setEnhanced(settings)      settings->flags |= FLAG_ENHANCEDREPORT
 #define setServerModeTime(settings)      settings->flags |= FLAG_SERVERMODETIME
-#define setPeerVerDetect(settings)      settings->flags_extend |= FLAG_PEERVER
+#define setPeerVerDetect(settings) settings->flags_extend |= FLAG_PEERVER
 #define setSeqNo64b(settings)      settings->flags_extend |= FLAG_SEQNO64
-#define setReverse(settings)      settings->flags_extend |= FLAG_REVERSE
+#define setReverse(settings)       settings->flags_extend |= FLAG_REVERSE
+#define setIsochronous(settings)   settings->flags_extend |= FLAG_ISOCHRONOUS
+#define setUDPTriggers(settings)   settings->flags_extend |= FLAG_UDPTRIGGERS
+#define setUDPHistogram(settings)  settings->flags_extend |= FLAG_UDPHISTOGRAM
+#define setL2LengthCheck(settings)    settings->flags_extend |= FLAG_L2LENGTHCHECK
+#define setIncrDstIP(settings)     settings->flags_extend |= FLAG_INCRDSTIP
+#define setTxSync(settings)        settings->flags_extend |= FLAG_TXSYNC
+#define setVaryLoad(settings)      settings->flags_extend |= FLAG_VARYLOAD
 
 #define unsetBuflenSet(settings)   settings->flags &= ~FLAG_BUFLENSET
 #define unsetCompat(settings)      settings->flags &= ~FLAG_COMPAT
@@ -320,6 +374,7 @@ typedef struct thread_Settings {
 #define unsetModeTime(settings)    settings->flags &= ~FLAG_MODETIME
 #define unsetReport(settings)      settings->flags &= ~FLAG_REPORTSETTINGS
 #define unsetMulticast(settings)   settings->flags &= ~FLAG_MULTICAST
+#define unsetSSMMulticast(settings)   settings->flags &= ~FLAG_SSM_MULTICAST
 #define unsetNoSettReport(settings)   settings->flags &= ~FLAG_NOSETTREPORT
 #define unsetNoConnReport(settings)   settings->flags &= ~FLAG_NOCONNREPORT
 #define unsetNoDataReport(settings)   settings->flags &= ~FLAG_NODATAREPORT
@@ -334,15 +389,31 @@ typedef struct thread_Settings {
 #define unsetServerModeTime(settings)    settings->flags &= ~FLAG_SERVERMODETIME
 #define unsetPeerVerDetect(settings)    settings->flags_extend &= ~FLAG_PEERVER
 #define unsetSeqNo64b(settings)    settings->flags_extend &= ~FLAG_SEQNO64
-#define unsetReverse(settings)    settings->flags_extend &= ~FLAG_REVERSE
+#define unsetReverse(settings)     settings->flags_extend &= ~FLAG_REVERSE
+#define unsetIsochronous(settings) settings->flags_extend &= ~FLAG_ISOCHRONOUS
+#define unsetUDPTriggers(settings) settings->flags_extend &= ~FLAG_UDPTRIGGERS
+#define unsetUDPHistogram(settings)    settings->flags_extend &= ~FLAG_UDPHISTOGRAM
+#define unsetL2LengthCheck(settings)  settings->flags_extend &= ~FLAG_L2LENGTHCHECK
+#define unsetIncrDstIP(settings)   settings->flags_extend &= ~FLAG_INCRDSTIP
+#define unsetTxSync(settings)      settings->flags_extend &= ~FLAG_TXSYNC
+#define unsetVaryLoad(settings)      settings->flags_extend &= ~FLAG_VARYLOAD
 
 /*
- * Messasge header flags
+ * Message header flags
  *
  * base flags, keep compatible with older versions
  */
 #define HEADER_VERSION1 0x80000000
 #define HEADER_EXTEND   0x40000000
+#define HEADER_UDPTESTS 0x20000000
+
+// Below flags are used to pass test settings in *every* UDP packet
+// and not just during the header exchange
+#define HEADER_UDP_ISOCH    0x00000001
+#define HEADER_L2ETHPIPV6   0x00000002
+#define HEADER_L2LENCHECK   0x00000004
+#define HEADER_UDPTRIGGERS  0x00000008
+
 #define RUN_NOW         0x00000001
 // newer flags
 #define UNITS_PPS             0x00000001
@@ -350,8 +421,12 @@ typedef struct thread_Settings {
 #define REALTIME              0x00000004
 #define REVERSE               0x00000008
 
+// later features
 #define HDRXACKMAX 2500000 // default 2.5 seconds, units microseconds
 #define HDRXACKMIN   10000 // default 10 ms, units microseconds
+#define MAGIC_NUMBER_TYPE 0x100 //
+#define MAGIC_NUMBER_LEN   0x4
+#define MAGIC_DHDHOST_TIMESTAMP 0x18EF0001 // Value used in the UDP payload to trigger the driver to insert DHD host timestamp
 
 /*
  * Structures used for test messages which
@@ -384,14 +459,13 @@ typedef struct UDP_datagram {
     unsigned int tv_sec  : 32;
     unsigned int tv_usec : 32;
 #endif //32
-#ifdef HAVE_SEQNO64b
 #ifdef HAVE_INT32_T
     u_int32_t id2;
 #else
     unsigned int id2      : 32;
 #endif // 32
-#endif //SEQNO64b
 } UDP_datagram;
+
 typedef struct hdr_typelen {
 #ifdef HAVE_INT32_T
     int32_t type;
@@ -402,6 +476,7 @@ typedef struct hdr_typelen {
 #endif
 } hdr_typelen;
 
+
 /*
  * The client_hdr structure is sent from clients
  * to servers to alert them of things that need
@@ -439,6 +514,8 @@ typedef struct client_hdr_v1 {
 #endif
 } client_hdr_v1;
 
+// This is used for tests that require
+// the initial handshake
 typedef struct client_hdrext {
     hdr_typelen typelen;
 #ifdef HAVE_INT32_T
@@ -460,6 +537,201 @@ typedef struct client_hdrext {
 #endif
 } client_hdrext;
 
+
+/*
+ * Isoch payload structure
+ *
+ *                 0      7 8     15 16    23 24    31
+ *                +--------+--------+--------+--------+
+ *      0x00  1   |          seqno lower              |
+ *                +--------+--------+--------+--------+
+ *      0x04  2   |             tv_sec                |
+ *                +--------+--------+--------+--------+
+ *      0x08  3   |             tv_usec               |
+ *                +--------+--------+--------+--------+
+ *      0x0c  4   |    (reserved) seqno upper         |
+ *                +--------+--------+--------+--------+
+ *            5   |         v1 hdr                    |
+ *                +--------+--------+--------+--------+
+ *            6   |         v1 hdr (continued)        |
+ *                +--------+--------+--------+--------+
+ *            7   |         v1 hdr (continued)        |
+ *                +--------+--------+--------+--------+
+ *            8   |         v1 hdr (continued)        |
+ *                +--------+--------+--------+--------+
+ *            9   |         v1 hdr (continued)        |
+ *                +--------+--------+--------+--------+
+ *            10  |         v1 hdr (final)            |
+ *                +--------+--------+--------+--------+
+ *            11  | udp test flags  | tlv offset      |
+ *                +--------+--------+--------+--------+
+ *            12  |        iperf version major        |
+ *                +--------+--------+--------+--------+
+ *            13  |        iperf version minor        |
+ *                +--------+--------+--------+--------+
+ *            14  |        ref sync sample tv_sec     |
+ *                +--------+--------+--------+--------+
+ *            15  |        ref sync sample tv_used    |
+ *                +--------+--------+--------+--------+
+ *            16  |        gps sync sample tv_sec     |
+ *                +--------+--------+--------+--------+
+ *            17  |        gps sync sample tv_usec    |
+ *                +--------+--------+--------+--------+
+ *            18  |        isoch burst period (us)    |
+ *                +--------+--------+--------+--------+
+ *            19  |        isoch start timestamp (s)  |
+ *                +--------+--------+--------+--------+
+ *            20  |        isoch start timestamp (us) |
+ *                +--------+--------+--------+--------+
+ *            21  |        isoch prev frameid         |
+ *                +--------+--------+--------+--------+
+ *            22  |        isoch frameid              |
+ *                +--------+--------+--------+--------+
+ *            23  |        isoch burtsize             |
+ *                +--------+--------+--------+--------+
+ *            24  |        isoch bytes remaining      |
+ *                +--------+--------+--------+--------+
+ *            25  |        isoch reserved             |
+ *                +--------+--------+--------+--------+
+ *            26  |        hw timestamps ...          |
+ *                +--------+--------+--------+--------+
+ *            n   |        hw timestamps ...
+ *                +--------+--------+--------+--------+
+ *
+ */
+
+typedef struct UDP_isoch_payload {
+#ifdef HAVE_INT32_T
+    u_int32_t burstperiod; //period units microseconds
+    u_int32_t start_tv_sec;
+    u_int32_t start_tv_usec;
+    u_int32_t prevframeid;
+    u_int32_t frameid;
+    u_int32_t burstsize;
+    u_int32_t remaining;
+    u_int32_t resevered;
+#else
+    unsigned int burstperiod : 32;
+    unsigned int start_tv_sec : 32;
+    unsigned int start_tv_usec : 32;
+    unsigned int prevframeid : 32;
+    unsigned int frameid : 32;
+    unsigned int burstsize : 32;
+    unsigned int remaining : 32;
+    unsigned int reserved : 32;
+#endif
+} UDP_isoch_payload;
+
+// This is used for UDP tests that don't
+// require any handshake, i.e they are stateless
+typedef struct client_hdr_udp_tests {
+// for 32 bit systems, skip over this field
+// so it remains interoperable with 64 bit peers
+#ifdef HAVE_INT32_T
+    u_int16_t testflags;
+    u_int16_t tlvoffset;
+    u_int32_t version_u;
+    u_int32_t version_l;
+    u_int32_t ref_sync_tv_sec;
+    u_int32_t ref_sync_tv_nsec;
+    u_int32_t gps_sync_tv_sec;
+    u_int32_t gps_sync_tv_nsec;
+#else
+    unsigned short testflags   : 16;
+    unsigned short tlvoffset   : 16;
+    unsigned int version_u   : 32;
+    unsigned int version_l   : 32;
+    unsigned int ref_sync_tv_sec : 32;
+    unsigned int ref_sync_tv_nsec : 32;
+    unsigned int gps_sync_tv_sec : 32;
+    unsigned int gps_sync_tv_nsec : 32;
+#endif
+} client_hdr_udp_tests;
+
+#ifdef HAVE_UDPTRIGGERS
+/*
+ *   UDP FW inline timestamps
+ *
+ *                 0      7 8     15 16    23 24    31
+ *                +--------+--------+--------+--------+
+ *            1   |  type (0x100)  |   length (68)    |   length including the 4 bytes tlv
+ *                +--------+--------+--------+--------+
+ *            2   |             host tx tv sec        |
+ *                +--------+--------+--------+--------+
+ *            3   |             host tx tv usec       |
+ *                +--------+--------+--------+--------+
+ *            4   |             host rx tv sec        |
+ *                +--------+--------+--------+--------+
+ *            5   |             host rx tv usec       |
+ *                +--------+--------+--------+--------+
+ *            6   |     type (0x1)  |      cnt (2)    |    fw rx tlv
+ *                +--------+--------+--------+--------+
+ *            7   |        fw rx ts 1 (mac)           |
+ *                +--------+--------+--------+--------+
+ *            8   |        fw rx ts 2 (pcie)          |
+ *                +--------+--------+--------+--------+
+ *            9   |     type (0x2)  |      cnt (1)    |    fw tx tlv
+ *                +--------+--------+--------+--------+
+ *            10  |        seqno lower                |
+ *                +--------+--------+--------+--------+
+ *            11  |        iperf tv_sec               |
+ *                +--------+--------+--------+--------+
+ *            12  |        iperf tv_usec              |
+ *                +--------+--------+--------+--------+
+ *            13  |        seqno upper ??             |
+ *                +--------+--------+--------+--------+
+ *            14  |        fw tx ts 1  pcie           |
+ *                +--------+--------+--------+--------+
+ *            15  |        fw tx ts 2  tx dma         |
+ *                +--------+--------+--------+--------+
+ *            16  |        fw tx ts 3  tx status      |
+ *                +--------+--------+--------+--------+
+ *            17  |        fw tx ts 4  pcie rt        |
+ *                +--------+--------+--------+--------+
+ *
+ * TSF histograms
+ * DHDRX-DHDTX 'DRRx-DRTx;'
+ * hs1 = 14,8 T6 "FWR2-FWT1"
+ * hs2 = 15,7 T5 "FWR1-FWT2'
+ * hs3 = 14,17 T4 'FWT4-FWT1'
+ * hs4 = 15,16 T3 'FWT3-FWT2'
+ * hs5 = 7,8 T2 'FWR2-FWR1'
+ */
+typedef struct fwtsfrx_t {
+    u_int16_t fwtype;
+    u_int16_t fwnct;
+    u_int32_t tsf_rxmac;
+    u_int32_t tsf_rxpcie;
+} fwtsfrx_t;
+
+typedef struct fwtsftx_t {
+    struct UDP_datagram udpid;
+    u_int32_t tsf_txpcie;
+    u_int32_t tsf_txdma;
+    u_int32_t tsf_txstatus;
+    u_int32_t tsf_txpciert;
+} fwtsftx_t;
+
+typedef struct UDPTriggers {
+    u_int16_t type;
+    u_int16_t length;
+    u_int32_t hosttx_tv_sec;
+    u_int32_t hosttx_tv_usec;
+    u_int32_t hostrx_tv_sec;
+    u_int32_t hostrx_tv_usec;
+    struct fwtsfrx_t fwtsf_rx;
+    u_int16_t fwtsf_type;
+    u_int16_t fwtsf_cnt;
+    struct fwtsftx_t fwtsf_tx[];
+} UDPTriggers;
+#endif //UDPTRIGGERS
+
+
+typedef struct client_hdr_udp_isoch_tests {
+    client_hdr_udp_tests udptests;
+    UDP_isoch_payload isoch;
+} client_hdr_udp_isoch_tests;
+
 typedef struct client_hdr_ack {
     hdr_typelen typelen;
 #ifdef HAVE_INT32_T
@@ -478,7 +750,10 @@ typedef struct client_hdr_ack {
 
 typedef struct client_hdr {
     client_hdr_v1 base;
-    client_hdrext extend;
+    union {
+	client_hdrext extend;
+	client_hdr_udp_tests udp;
+    };
 } client_hdr;
 
 /*
