@@ -81,7 +81,11 @@
 #include "util.h"
 #include "version.h"
 #include "Locale.h"
+#include "SocketAddr.h"
 
+#ifdef HAVE_SSM_MULTICAST
+#include <net/if.h>
+#endif
 /* -------------------------------------------------------------------
  * Stores local hostname and socket info.
  * ------------------------------------------------------------------- */
@@ -99,7 +103,7 @@ Listener::Listener( thread_Settings *inSettings ) {
      */
     mSettings = inSettings;
 
-    // initialize buffer for packets
+    // alloc and initialize the buffer (mBuf) used for packet reads()
     mBuf = new char[((mSettings->mBufLen > SIZEOF_MAXHDRMSG) ? mSettings->mBufLen : SIZEOF_MAXHDRMSG)];
     FAIL_errno( mBuf == NULL, "No memory for buffer\n", mSettings );
     /*
@@ -175,7 +179,6 @@ void Listener::Run( void ) {
         Settings_Copy( mSettings, &server );
         server->mThreadMode = kMode_Server;
 
-
         // Accept each packet,
         // If there is no existing client, then start
         // a new thread to service the new client
@@ -202,6 +205,7 @@ sInterupted == SIGALRM
 		    break;
 		}
             }
+
             // Reset Single Client Stuff
             if ( isSingleClient( mSettings ) && clients == NULL ) {
                 mSettings->peer = server->peer;
@@ -229,8 +233,9 @@ sInterupted == SIGALRM
                 }
             }
             // Check for exchange of test information and also determine v2.0.5 vs 2.0.10+
-            if ( !isCompat( mSettings ) && !isMulticast( mSettings ) ) {
-                if (ReadClientHeader(hdr) < 0) {
+            if ( !isCompat( mSettings )) {
+		int flags;
+                if ((flags = ReadClientHeader(hdr)) < 0) {
 		    close( server->mSock );
 		    continue;
 		}
@@ -274,13 +279,25 @@ sInterupted == SIGALRM
                 Mutex_Unlock( &groupCond );
             }
 
+
+	    // Perform L2 setup if needed
+	    if (isUDP(mSettings) && (isL2LengthCheck(mSettings) || isL2LengthCheck(server))) {
+		if (L2_setup() < 0) {
+		    // L2 not allowed, abort this server try
+		    delete server;
+		    mSettings->mSock = -1;
+		}
+	    }
             // Store entry in connection list
-            Iperf_pushback( listtemp, &clients );
+	    if (mSettings->mSock > 0)
+		Iperf_pushback( listtemp, &clients );
             Mutex_Unlock( &clients_mutex );
 
-            // Start the server
+            //
+	    // Everything is now ready to start the server
+	    //
 #if defined(WIN32) && defined(HAVE_THREAD)
-            if ( UDP ) {
+            if ( UDP && mSettings->mSock > 0) {
                 // WIN32 does bad UDP handling so run single threaded
                 if ( server->runNow != NULL ) {
                     thread_start( server->runNow );
@@ -291,9 +308,12 @@ sInterupted == SIGALRM
                 }
             } else
 #endif
-            thread_start( server );
-
-            // create a new socket
+		{
+		    if (mSettings->mSock > 0)
+			thread_start( server );
+		}
+	    // create a new socket for the Listener thread now that server thread
+	    // is handling the current one
             if ( UDP ) {
                 mSettings->mSock = -1;
                 Listen( );
@@ -324,29 +344,30 @@ void Listener::Listen( ) {
 
     SockAddr_localAddr( mSettings );
 
-    // create an internet TCP socket
+    // create an AF_INET socket for the accepts
+    // for the case of L2 testing and UDP, a new AF_PACKET
+    // will be created to supercede this one
     int type = (isUDP( mSettings )  ?  SOCK_DGRAM  :  SOCK_STREAM);
     int domain = (SockAddr_isIPv6( &mSettings->local ) ?
 #ifdef HAVE_IPV6
-                  AF_INET6
+		  AF_INET6
 #else
-                  AF_INET
+		  AF_INET
 #endif
-                  : AF_INET);
+		  : AF_INET);
 
 #ifdef WIN32
     if ( SockAddr_isMulticast( &mSettings->local ) ) {
-        // Multicast on Win32 requires special handling
-        mSettings->mSock = WSASocket( domain, type, 0, 0, 0, WSA_FLAG_MULTIPOINT_C_LEAF | WSA_FLAG_MULTIPOINT_D_LEAF );
-        WARN_errno( mSettings->mSock == INVALID_SOCKET, "socket" );
+	// Multicast on Win32 requires special handling
+	mSettings->mSock = WSASocket( domain, type, 0, 0, 0, WSA_FLAG_MULTIPOINT_C_LEAF | WSA_FLAG_MULTIPOINT_D_LEAF );
+	WARN_errno( mSettings->mSock == INVALID_SOCKET, "socket" );
 
     } else
 #endif
-    {
-        mSettings->mSock = socket( domain, type, 0 );
-        WARN_errno( mSettings->mSock == INVALID_SOCKET, "socket" );
-    }
-
+	{
+	    mSettings->mSock = socket( domain, type, 0 );
+	    WARN_errno( mSettings->mSock == INVALID_SOCKET, "socket" );
+	}
     SetSocketOptions( mSettings );
 
     // reuse the address, so we can run if a former server was killed off
@@ -357,66 +378,322 @@ void Listener::Listen( ) {
     // bind socket to server address
 #ifdef WIN32
     if ( SockAddr_isMulticast( &mSettings->local ) ) {
-        // Multicast on Win32 requires special handling
-        rc = WSAJoinLeaf( mSettings->mSock, (sockaddr*) &mSettings->local, mSettings->size_local,0,0,0,0,JL_BOTH);
-        WARN_errno( rc == SOCKET_ERROR, "WSAJoinLeaf (aka bind)" );
+	// Multicast on Win32 requires special handling
+	rc = WSAJoinLeaf( mSettings->mSock, (sockaddr*) &mSettings->local, mSettings->size_local,0,0,0,0,JL_BOTH);
+	WARN_errno( rc == SOCKET_ERROR, "WSAJoinLeaf (aka bind)" );
     } else
 #endif
-    {
-        rc = bind( mSettings->mSock, (sockaddr*) &mSettings->local, mSettings->size_local );
-        FAIL_errno( rc == SOCKET_ERROR, "bind", mSettings );
-    }
+	{
+	    rc = bind( mSettings->mSock, (sockaddr*) &mSettings->local, mSettings->size_local );
+	    FAIL_errno( rc == SOCKET_ERROR, "bind", mSettings );
+	}
     // listen for connections (TCP only).
     // default backlog traditionally 5
     if ( !isUDP( mSettings ) ) {
-        rc = listen( mSettings->mSock, 5 );
-        WARN_errno( rc == SOCKET_ERROR, "listen" );
+	rc = listen( mSettings->mSock, 5 );
+	WARN_errno( rc == SOCKET_ERROR, "listen" );
     }
 
 #ifndef WIN32
     // if multicast, join the group
     if ( SockAddr_isMulticast( &mSettings->local ) ) {
-        McastJoin( );
+#ifdef HAVE_MULTICAST
+	McastJoin( );
+#else
+	fprintf(stderr, "Multicast not supported");
+#endif // HAVE_MULTICAST
     }
 #endif
 } // end Listen
 
 /* -------------------------------------------------------------------
- * Joins the multicast group, with the default interface.
+ * Joins the multicast group or source and group (SSM S,G)
+ *
+ * taken from: https://www.ibm.com/support/knowledgecenter/en/SSLTBW_2.1.0/com.ibm.zos.v2r1.hale001/ipv6d0141001708.htm
+ *
+ * Multicast function	                                        IPv4	                   IPv6	                Protocol-independent
+ * ==================                                           ====                       ====                 ====================
+ * Level of specified option on setsockopt()/getsockopt()	IPPROTO_IP	           IPPROTO_IPV6	IPPROTO_IP or IPPROTO_IPV6
+ * Join a multicast group	                                IP_ADD_MEMBERSHIP          IPV6_JOIN_GROUP	MCAST_JOIN_GROUP
+ * Leave a multicast group or leave all sources of that
+ *   multicast group	                                        IP_DROP_MEMBERSHIP	   IPV6_LEAVE_GROUP	MCAST_LEAVE_GROUP
+ * Select outbound interface for sending multicast datagrams	IP_MULTICAST_IF	IPV6_MULTICAST_IF	NA
+ * Set maximum hop count	                                IP_MULTICAST_TTL	   IPV6_MULTICAST_HOPS	NA
+ * Enable multicast loopback	                                IP_MULTICAST_LOOP	   IPV6_MULTICAST_LOOP	NA
+ * Join a source multicast group	                        IP_ADD_SOURCE_MEMBERSHIP   NA	                MCAST_JOIN_SOURCE_GROUP
+ * Leave a source multicast group	                        IP_DROP_SOURCE_MEMBERSHIP  NA	                MCAST_LEAVE_SOURCE_GROUP
+ * Block data from a source to a multicast group	        IP_BLOCK_SOURCE   	   NA	                MCAST_BLOCK_SOURCE
+ * Unblock a previously blocked source for a multicast group	IP_UNBLOCK_SOURCE	   NA	                MCAST_UNBLOCK_SOURCE
+ *
+ *
+ * Reminder:  The os will decide which version of IGMP or MLD to use.   This may be controlled by system settings, e.g.:
+ *
+ * [rmcmahon@lvnvdb0987:~/Code/ssm/iperf2-code] $ sysctl -a | grep mld | grep force
+ * net.ipv6.conf.all.force_mld_version = 0
+ * net.ipv6.conf.default.force_mld_version = 0
+ * net.ipv6.conf.lo.force_mld_version = 0
+ * net.ipv6.conf.eth0.force_mld_version = 0
+ *
+ * [rmcmahon@lvnvdb0987:~/Code/ssm/iperf2-code] $ sysctl -a | grep igmp | grep force
+ * net.ipv4.conf.all.force_igmp_version = 0
+ * net.ipv4.conf.default.force_igmp_version = 0
+ * net.ipv4.conf.lo.force_igmp_version = 0
+ * net.ipv4.conf.eth0.force_igmp_version = 0
+ *
  * ------------------------------------------------------------------- */
 
 void Listener::McastJoin( ) {
-#ifdef HAVE_MULTICAST
-    if ( !SockAddr_isIPv6( &mSettings->local ) ) {
-        struct ip_mreq mreq;
+    // This is the older mulitcast join code.  Both SSM and binding the
+    // an interface requires the newer socket options.  Using the older
+    // code here will maintain compatiblity with previous iperf versions
+    if (!isSSMMulticast(mSettings) && !mSettings->mIfrname) {
+	if ( !SockAddr_isIPv6( &mSettings->local ) ) {
 
-        memcpy( &mreq.imr_multiaddr, SockAddr_get_in_addr( &mSettings->local ),
-                sizeof(mreq.imr_multiaddr));
+	    struct ip_mreq mreq;
 
-        mreq.imr_interface.s_addr = htonl( INADDR_ANY );
+	    memcpy( &mreq.imr_multiaddr, SockAddr_get_in_addr( &mSettings->local ),
+		    sizeof(mreq.imr_multiaddr));
 
-        int rc = setsockopt( mSettings->mSock, IPPROTO_IP, IP_ADD_MEMBERSHIP,
-                             (char*) &mreq, sizeof(mreq));
-        WARN_errno( rc == SOCKET_ERROR, "multicast join" );
-    }
+	    mreq.imr_interface.s_addr = htonl( INADDR_ANY );
+
+	    int rc = setsockopt( mSettings->mSock, IPPROTO_IP, IP_ADD_MEMBERSHIP,
+				 (char*) &mreq, sizeof(mreq));
+	    WARN_errno( rc == SOCKET_ERROR, "multicast join" );
+
+	} else {
 #ifdef HAVE_IPV6_MULTICAST
-      else {
-        struct ipv6_mreq mreq;
+	    struct ipv6_mreq mreq;
 
-        memcpy( &mreq.ipv6mr_multiaddr, SockAddr_get_in6_addr( &mSettings->local ),
-                sizeof(mreq.ipv6mr_multiaddr));
+	    memcpy( &mreq.ipv6mr_multiaddr, SockAddr_get_in6_addr( &mSettings->local ),
+		    sizeof(mreq.ipv6mr_multiaddr));
 
-        mreq.ipv6mr_interface = 0;
+	    mreq.ipv6mr_interface = 0;
 
-        int rc = setsockopt( mSettings->mSock, IPPROTO_IPV6, IPV6_ADD_MEMBERSHIP,
-                             (char*) &mreq, sizeof(mreq));
-        WARN_errno( rc == SOCKET_ERROR, "multicast join" );
+	    int rc = setsockopt( mSettings->mSock, IPPROTO_IPV6, IPV6_ADD_MEMBERSHIP,
+				 (char*) &mreq, sizeof(mreq));
+	    WARN_errno( rc == SOCKET_ERROR, "multicast v6 join" );
+#else
+	    fprintf(stderr, "Unfortunately, IPv6 multicast is not supported on this platform\n");
+#endif
+	}
+    } else {
+#ifdef HAVE_SSM_MULTICAST
+	// Here it's either an SSM S,G multicast join or a *,G with an interface specifier
+	// Use the newer socket options when these are specified
+	socklen_t socklen = sizeof(struct sockaddr_storage);
+	int iface=0;
+	int rc;
+
+	/* Set the interface or any */
+	if (mSettings->mIfrname) {
+	    iface = if_nametoindex(mSettings->mIfrname);
+	    FAIL_errno(!iface, "mcast if_nametoindex",mSettings);
+	} else {
+	    iface = 0;
+	}
+
+        if (isIPV6(mSettings)) {
+#ifdef HAVE_IPV6_MULTICAST
+	    if (mSettings->mSSMMulticastStr) {
+		struct group_source_req group_source_req;
+		struct sockaddr_in6 *group;
+		struct sockaddr_in6 *source;
+
+		group_source_req.gsr_interface = iface;
+		group=(struct sockaddr_in6*)&group_source_req.gsr_group;
+		source=(struct sockaddr_in6*)&group_source_req.gsr_source;
+		source->sin6_family = AF_INET6;
+		group->sin6_family = AF_INET6;
+		/* Set the group */
+		rc=getsockname(mSettings->mSock,(struct sockaddr *)group, &socklen);
+		FAIL_errno( rc == SOCKET_ERROR, "mcast join source group getsockname",mSettings );
+		group->sin6_port = 0;    /* Ignored */
+
+		/* Set the source, apply the S,G */
+		rc=inet_pton(AF_INET6, mSettings->mSSMMulticastStr,&source->sin6_addr);
+		FAIL_errno( rc != 1, "mcast v6 join source group pton",mSettings );
+		source->sin6_port = 0;    /* Ignored */
+#ifdef HAVE_STRUCT_SOCKADDR_IN_SIN_LEN
+		source->sin_len = group->sin_len;
+#endif
+		rc = setsockopt(mSettings->mSock,IPPROTO_IPV6,MCAST_JOIN_SOURCE_GROUP, &group_source_req,
+			    sizeof(group_source_req));
+		FAIL_errno( rc == SOCKET_ERROR, "mcast v6 join source group",mSettings);
+	    } else {
+		struct group_req group_req;
+		struct sockaddr_in6 *group;
+
+		group_req.gr_interface = iface;
+		group=(struct sockaddr_in6*)&group_req.gr_group;
+		group->sin6_family = AF_INET6;
+		/* Set the group */
+		rc=getsockname(mSettings->mSock,(struct sockaddr *)group, &socklen);
+		FAIL_errno( rc == SOCKET_ERROR, "mcast v6 join group getsockname",mSettings );
+		group->sin6_port = 0;    /* Ignored */
+		rc = setsockopt(mSettings->mSock,IPPROTO_IPV6,MCAST_JOIN_GROUP, &group_req,
+				sizeof(group_source_req));
+		FAIL_errno( rc == SOCKET_ERROR, "mcast v6 join group",mSettings);
+	    }
+#else
+	    fprintf(stderr, "Unfortunately, IPv6 multicast is not supported on this platform\n");
+#endif
+	} else {
+	    if (mSettings->mSSMMulticastStr) {
+		struct group_source_req group_source_req;
+		struct sockaddr_in *group;
+
+		struct sockaddr_in *source;
+		group_source_req.gsr_interface = iface;
+		group=(struct sockaddr_in*)&group_source_req.gsr_group;
+		source=(struct sockaddr_in*)&group_source_req.gsr_source;
+		source->sin_family = AF_INET;
+		group->sin_family = AF_INET;
+		/* Set the group */
+		rc=getsockname(mSettings->mSock,(struct sockaddr *)group, &socklen);
+		FAIL_errno( rc == SOCKET_ERROR, "mcast join source group getsockname",mSettings );
+		group->sin_port = 0;    /* Ignored */
+
+		/* Set the source, apply the S,G */
+		rc=inet_pton(AF_INET,mSettings->mSSMMulticastStr,&source->sin_addr);
+		FAIL_errno(rc != 1, "mcast join source pton",mSettings );
+#ifdef HAVE_STRUCT_SOCKADDR_IN_SIN_LEN
+		source->sin_len = group->sin_len;
+#endif
+		source->sin_port = 0;    /* Ignored */
+		rc = setsockopt(mSettings->mSock,IPPROTO_IP,MCAST_JOIN_SOURCE_GROUP, &group_source_req,
+				sizeof(group_source_req));
+		FAIL_errno( rc == SOCKET_ERROR, "mcast join source group",mSettings);
+	    } else {
+		struct group_req group_req;
+		struct sockaddr_in *group;
+
+		group_req.gr_interface = iface;
+		group=(struct sockaddr_in*)&group_req.gr_group;
+		group->sin_family = AF_INET;
+		/* Set the group */
+		rc=getsockname(mSettings->mSock,(struct sockaddr *)group, &socklen);
+		FAIL_errno( rc == SOCKET_ERROR, "mcast join group getsockname",mSettings );
+		group->sin_port = 0;    /* Ignored */
+		rc = setsockopt(mSettings->mSock,IPPROTO_IP,MCAST_JOIN_GROUP, &group_req,
+				sizeof(group_source_req));
+		FAIL_errno( rc == SOCKET_ERROR, "mcast join group",mSettings);
+	    }
+	}
+#else
+	fprintf(stderr, "Unfortunately, SSM is not supported on this platform\n");
+	exit(-1);
+#endif
     }
-#endif
-#endif
 }
 // end McastJoin
 
+int Listener::L2_setup (void) {
+#if defined(HAVE_LINUX_FILTER_H) && defined(HAVE_AF_PACKET)
+    //
+    //  Supporting parallel L2 UDP threads is a bit tricky.  Below are some notes as to why and the approach used.
+    //
+    //  The primary issues for UDP are:
+    //
+    //  1) We want the listener thread to hand off the flow to a server thread and not be burdened by that flow
+    //  2) For -P support, the listener thread neads to detect new flows which will share the same UDP port
+    //     and UDP is stateless
+    //
+    //  The listener thread needs to detect new traffic flows and hand them to a new server thread, and then
+    //  rehang a listen/accept.  For standard iperf the "flow routing" is done using connect() per the ip quintuple.
+    //  The OS will then route established connected flows to the socket descriptor handled by a server thread and won't
+    //  burden the listener thread with these packets.
+    //
+    //  For L2 verification, we have to create a two sockets that will exist for the life of the flow.  A
+    //  new packet socket (AF_PACKET) will receive L2 frames and bypasses
+    //  the OS network stack.  The original AF_INET socket will still send up packets
+    //  to the network stack.
+    //
+    //  When using packet sockets there is inherent packet duplication, the hand off to a server
+    //  thread is not so straight forward as packets will continue being sent up to the listener thread
+    //  (technical problem is that packet sockets do not support connect() which binds the IP quintuple as the
+    //  forwarding key) Since the Listener uses recvfrom(), there is no OS mechanism to detect new flows nor
+    //  to drop packets.  The listener can't listen on quintuple based connected flows because the client's source
+    //  port is unknown.  Therefore the Listener thread will continue to receive packets from all established
+    //  flows sharing the same dst port which will impact CPU utilization and hence performance.
+    //
+    //  The technique used to address this is to open an AF_PACKET socket and leave the AF_INET socket open.
+    //  (This also aligns with BSD based systems)  The original AF_INET socket will remain in the (connected)
+    //  state so the network stack has it's connected state.  A cBPF is then used to cause the kernel to fast drop
+    //  those packets.  A cBPF is set up to drop such packets.  The test traffic will then only come over the
+    //  packet (raw) socket and not the  AF_INET socket. If we were to try to close the original AF_INET socket
+    //  (vs leave it open w/the fast drop cBPF) then the existing traffic will be sent up by the network stack
+    //  to he Listener thread, flooding it with packets, again something we want to avoid.
+    //
+    //  On the packet (raw) socket itself, we do two more things to better handle performance
+    //
+    //  1)  Use a full quintuple cBPF allowing the kernel to filter packets (allow) per the quintuple
+    //  2)  Use the packet fanout option to assign a CBPF to a socket and hence to a single server thread minimizing
+    //      duplication (reduce all cBPF's filtering load)
+    //
+    struct sockaddr *p = (sockaddr *)&server->peer;
+    struct sockaddr *l = (sockaddr *)&server->local;
+    int rc = 0;
+
+    //
+    // Establish a packet (raw) socket to be used by the server thread giving it full L2 packets
+    //
+    struct sockaddr s;
+    socklen_t len = sizeof(s);
+    getpeername(mSettings->mSock, &s, &len);
+    if (isIPV6(server)) {
+	server->mSock = socket(AF_PACKET, SOCK_RAW, htons(ETH_P_IPV6));
+	WARN_errno(server->mSock == INVALID_SOCKET, "ip6 packet socket (AF_PACKET)");
+	server->l4offset = IPV6HDRLEN + sizeof(struct ether_header);
+    } else {
+	server->mSock = socket(AF_PACKET, SOCK_RAW, htons(ETH_P_IP));
+	WARN_errno(server->mSock == INVALID_SOCKET, "ip packet socket (AF_PACKET)");
+	unsetIPV6(server);
+	server->l4offset = sizeof(struct iphdr) + sizeof(struct ether_header);
+    }
+    // Didn't get a valid socket, return now
+    if (server->mSock < 0) {
+	return server->mSock;
+    }
+    // More per thread settings based on using a packet socket
+    server->l4payloadoffset = server->l4offset + sizeof(struct udphdr);
+    server->recvflags = MSG_TRUNC;
+    // The original AF_INET socket only exists to keep the connected state
+    // in the OS for this flow. Fast drop packets there as
+    // now packets will use the AF_PACKET (raw) socket
+    // Also, store the original AF_INET socket descriptor so it can be
+    // closed in the Server's destructor.  (Note: closing the
+    // socket descriptors will also free the cBPF.)
+    //
+    server->mSockDrop = mSettings->mSock;
+    rc = SockAddr_Drop_All_BPF(mSettings->mSock);
+    WARN_errno( rc == SOCKET_ERROR, "l2 all drop bpf");
+
+    // Now optimize packet flow up the raw socket
+    // Establish the flow BPF to forward up only "connected" packets to this raw socket
+    if (l->sa_family == AF_INET6) {
+	struct in6_addr *v6peer = SockAddr_get_in6_addr(&server->peer);
+	struct in6_addr *v6local = SockAddr_get_in6_addr(&server->local);
+	if (isIPV6(server)) {
+	    rc = SockAddr_v6_Connect_BPF(server->mSock, v6local, v6peer, ((struct sockaddr_in6 *)(l))->sin6_port, ((struct sockaddr_in6 *)(p))->sin6_port);
+	    WARN_errno( rc == SOCKET_ERROR, "l2 connect ipv6 bpf");
+	} else {
+	    // This is an ipv4 address in a v6 family (structure), just pull the lower 32 bits for the v4 addr
+	    rc = SockAddr_v4_Connect_BPF(server->mSock, (uint32_t) v6local->s6_addr32[3], (uint32_t) v6peer->s6_addr32[3], ((struct sockaddr_in6 *)(l))->sin6_port, ((struct sockaddr_in6 *)(p))->sin6_port);
+	    WARN_errno( rc == SOCKET_ERROR, "l2 v4in6 connect ip bpf");
+	}
+    } else {
+	rc = SockAddr_v4_Connect_BPF(server->mSock, ((struct sockaddr_in *)(l))->sin_addr.s_addr, ((struct sockaddr_in *)(p))->sin_addr.s_addr, ((struct sockaddr_in *)(l))->sin_port, ((struct sockaddr_in *)(p))->sin_port);
+	WARN_errno( rc == SOCKET_ERROR, "l2 connect ip bpf");
+    }
+    if (rc < 0)
+	return -1;
+    else
+	return 1;
+#else
+    fprintf(stderr, "Client requested --l2checks but not supported on this platform\n");
+    return -1;
+#endif
+}
 /* -------------------------------------------------------------------
  * Sets the Multicast TTL for outgoing packets.
  * ------------------------------------------------------------------- */
@@ -478,29 +755,39 @@ void Listener::Accept( thread_Settings *server ) {
 	    }
 	}
 	if ( isUDP( server ) ) {
+	    int rc;
 	    /* ------------------------------------------------------------------------
 	     * Do the equivalent of an accept() call for UDP sockets. This waits
 	     * on a listening UDP socket until we get a datagram.
 	     * ------------------------------------------------------------------- ----*/
-	    int rc;
 	    Iperf_ListEntry *exist;
-	    int32_t datagramID;
+	    // Preset the server socket to INVALID, hang recvfrom on the Listener's socket
+	    // The INVALID socket is used to keep the while loop going
 	    server->mSock = INVALID_SOCKET;
-
 	    rc = recvfrom( mSettings->mSock, mBuf, mSettings->mBufLen, 0,
-			   (struct sockaddr*) &server->peer, &server->size_peer );
+			       (struct sockaddr*) &server->peer, &server->size_peer );
 	    FAIL_errno( rc == SOCKET_ERROR, "recvfrom", mSettings );
 	    Mutex_Lock( &clients_mutex );
 
 	    // Handle connection for UDP sockets.
 	    exist = Iperf_present( &server->peer, clients);
-	    datagramID = ntohl( ((UDP_datagram*) mBuf)->id );
-	    if ( exist == NULL && datagramID >= 0 ) {
+	    if ( exist == NULL ) {
+		// We have a new UDP flow so let's start the
+		// process to handle it and in a new server thread (yet to be created)
 		server->mSock = mSettings->mSock;
+		// This connect() will allow the OS to only
+		// send packets with the ip quintuple up to the server
+		// socket and, hence, to the server thread (yet to be created)
+		// This connect() routing is only supported with AF_INET or AF_INET6 sockets,
+		// e.g. AF_PACKET sockets can't do this.  We'll handle packet sockets later
+		// All UDP accepts here will use AF_INET.  This is intentional and needed
 		int rc = connect( server->mSock, (struct sockaddr*) &server->peer,
 				  server->size_peer );
 		FAIL_errno( rc == SOCKET_ERROR, "connect UDP", mSettings );
 	    } else {
+		// This isn't a new flow so just ignore the packet
+		// and continue with the while loop
+		// printf("Debug: drop packet on sock %d\n",mSettings->mSock);
 		server->mSock = INVALID_SOCKET;
 	    }
 	    Mutex_Unlock( &clients_mutex );
@@ -637,7 +924,11 @@ void Listener::UDPSingleServer( ) {
                         hdr = (server_hdr*) (UDP_Hdr+1);
 
                         hdr->base.flags        = htonl( HEADER_VERSION1 );
-                        hdr->base.total_len1   = htonl( (long) (stats->TotalLen >> 32) );
+#ifdef HAVE_QUAD_SUPPORT
+			hdr->base.total_len1   = htonl( (long) (stats->TotalLen >> 32) );
+#else
+			hdr->base.total_len1   = htonl(0x0);
+#endif
                         hdr->base.total_len2   = htonl( (long) (stats->TotalLen & 0xFFFFFFFF) );
                         hdr->base.stop_sec     = htonl( (long) stats->endTime );
                         hdr->base.stop_usec    = htonl( (long)((stats->endTime - (long)stats->endTime) \
@@ -758,8 +1049,33 @@ void Listener::UDPSingleServer( ) {
 
 int Listener::ReadClientHeader(client_hdr *hdr ) {
     int flags = 0;
+    int testflags = 0;
     if (isUDP(mSettings)) {
 	flags = ntohl(hdr->base.flags);
+	if ((flags & HEADER_UDPTESTS) != 0) {
+	    testflags = ntohs(hdr->udp.testflags);
+	    // Handle stateless flags
+#ifdef HAVE_ISOCHRONOUS
+	    if ((testflags & HEADER_UDP_ISOCH) != 0) {
+		setIsochronous(server);
+	    }
+#endif
+	    if ((testflags & HEADER_L2ETHPIPV6) != 0) {
+		setIPV6(server);
+	    } else {
+		unsetIPV6(server);
+	    }
+	    if ((testflags & HEADER_L2LENCHECK) != 0) {
+		setL2LengthCheck(server);
+	    }
+#ifdef HAVE_UDPTRIGGERS
+	    if ((testflags & HEADER_UDPTRIGGERS) != 0) {
+		setUDPTriggers(server);
+		setSeqNo64b(server);
+	    }
+#endif
+	    reporter_peerversion(server, ntohl(hdr->udp.version_u), ntohl(hdr->udp.version_l));
+	}
     } else {
 	int n, len;
 	char *p = (char *)hdr;
@@ -804,12 +1120,15 @@ int Listener::ReadClientHeader(client_hdr *hdr ) {
 	    }
 	}
     }
+    // Handle flags that require an ack back to the client
     if ((flags & HEADER_EXTEND) != 0 ) {
 	reporter_peerversion(server, ntohl(hdr->extend.version_u), ntohl(hdr->extend.version_l));
 	//  Extended header successfully read. Ack the client with our version info now
-	ClientHeaderAck();
+	if (!isMulticast(mSettings)) {
+	    ClientHeaderAck();
+	}
     }
-    return 1;
+    return(flags);
 }
 
 int Listener::ClientHeaderAck(void) {
