@@ -51,6 +51,7 @@
  * Strings and other stuff that is locale specific.
  * ------------------------------------------------------------------- */
 #include <inttypes.h>
+#include "headers.h"
 #include "version.h"
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -92,7 +93,7 @@ Client/Server:\n\
 #ifdef HAVE_SCHED_SETSCHEDULER
 "  -z, --realtime           request realtime scheduler\n"
 #endif
-"  -B, --bind      <host>   bind to <host>, an interface or multicast address\n\
+"  -B, --bind <host>[:<port>][%<dev>] bind to <host>, ip addr (including multicast address) and optional port and device\n\
   -C, --compatibility      for use with older versions does not sent extra msgs\n\
   -M, --mss       #        set TCP maximum segment size (MTU - 40 bytes)\n\
   -N, --nodelay            set TCP no delay, disabling Nagle's Algorithm\n\
@@ -101,6 +102,9 @@ Client/Server:\n\
 Server specific:\n\
   -s, --server             run in server mode\n\
   -t, --time      #        time in seconds to listen for new connections as well as to receive traffic (default not set)\n\
+      --udp-histogram #,#  enable UDP latency histogram(s) with bin width and count, e.g. 1,1000=1(ms),1000(bins)\n\
+  -B, --bind <ip>[%<dev>]  bind to multicast address and optional device\n\
+  -H, --ssm-host <ip>      set the SSM source, use with -B for (S,G) \n\
   -U, --single_udp         run in single threaded UDP mode\n\
   -D, --daemon             run the server as a daemon\n"
 #ifdef WIN32
@@ -113,11 +117,15 @@ const char usage_long2[] = "\
 \n\
 Client specific:\n\
   -c, --client    <host>   run in client mode, connecting to <host>\n\
-  -d, --dualtest           Do a bidirectional test simultaneously\n\
-  -n, --num       #[kmgKMG]    number of bytes to transmit (instead of -t)\n\
+  -d, --dualtest           Do a bidirectional test simultaneously\n"
+#ifdef HAVE_ISOCHRONOUS
+"      --ipg                set the the interpacket gap (milliseconds) for packets within an isochronous frame\n\
+      --isochronous <frames-per-second>:<mean>,<stddev> send traffic in bursts (frames - emulate video traffic)\n"
+#endif
+"  -n, --num       #[kmgKMG]    number of bytes to transmit (instead of -t)\n\
   -r, --tradeoff           Do a bidirectional test individually\n\
   -t, --time      #        time in seconds to transmit for (default 10 secs)\n\
-  -B, --bind [<ip> | <ip:port>] bind src addr(s) from which to originate traffic\n\
+  -B, --bind [<ip> | <ip:port>] bind ip (and optional port) from which to source traffic\n\
   -F, --fileinput <name>   input the data to be transmitted from a file\n\
   -I, --stdin              input the data to be transmitted from stdin\n\
   -L, --listenport #       port to receive bidirectional tests back on\n\
@@ -187,9 +195,21 @@ const char multicast_ttl[] =
 const char join_multicast[] =
 "Joining multicast group  %s\n";
 
+const char join_multicast_sg[] =
+"Joining multicast (S,G)=%s,%s\n";
+
+const char join_multicast_starg_dev[] =
+"Joining multicast (*,G)=*,%s w/iface %s\n";
+
+const char join_multicast_sg_dev[] =
+"Joining multicast (S,G)=%s,%s w/iface %s\n";
+
 const char client_datagram_size[] =
 "Sending %d byte datagrams, IPG target: %4.2f us\n";
 
+const char client_datagram_size_tx_sync[] =
+"Sending %d byte datagrams, tx-sync periodic: %4.2f ms\n";
+
 const char client_datagram_size_kalman[] =
 "Sending %d byte datagrams, IPG target: %4.2f us (kalman adjust)\n";
 
@@ -208,6 +228,9 @@ const char window_default[] =
 const char wait_server_threads[] =
 "Waiting for server threads to complete. Interrupt again to force quit.\n";
 
+const char client_udp_isochronous[] =
+"UDP isochronous: %d frames/sec mean=%s/s, stddev=%s/s, Period/IPG=%0.2f/%0.3f ms\n";
+
 /* -------------------------------------------------------------------
  * Legacy reports
  * ------------------------------------------------------------------- */
@@ -224,10 +247,10 @@ const char report_sum_bw_format[] =
 const char report_bw_jitter_loss_header[] =
 "[ ID] Interval       Transfer     Bandwidth        Jitter   Lost/Total Datagrams\n";
 const char report_bw_jitter_loss_format[] =
-"[%3d] %4.1f-%4.1f sec  %ss  %ss/sec  %6.3f ms %4"PRIdMAX"/%5"PRIdMAX" (%.2g%%)\n";
+"[%3d] %4.1f-%4.1f sec  %ss  %ss/sec  %6.3f ms %4" IPERFdMAX "/%5" IPERFdMAX " (%.2g%%)\n";
 
 const char report_sum_bw_jitter_loss_format[] =
-"[SUM] %4.1f-%4.1f sec  %ss  %ss/sec  %6.3f ms %4"PRIdMAX"/%5"PRIdMAX" (%.2g%%)\n";
+"[SUM] %4.1f-%4.1f sec  %ss  %ss/sec  %6.3f ms %4" IPERFdMAX "/%5" IPERFdMAX " (%.2g%%)\n";
 
 /* -------------------------------------------------------------------
  * Enhanced reports (per -e)
@@ -275,26 +298,39 @@ const char report_sum_bw_write_enhanced_format[] =
 #endif
 
 const char report_bw_pps_enhanced_header[] =
-"[ ID] Interval       Transfer     Bandwidth      PPS\n";
+"[ ID] Interval       Transfer     Bandwidth      Write/Err  PPS\n";
 
 const char report_bw_pps_enhanced_format[] =
-"[%3d] %4.2f-%4.2f sec  %ss  %ss/sec %4.0f pps\n";
+"[%3d] %4.2f-%4.2f sec  %ss  %ss/sec  %d/%d %8.0f pps\n";
+
+const char report_bw_pps_enhanced_isoch_header[] =
+"[ ID] Interval       Transfer     Bandwidth      Write/Err  PPS  frames:tx/missed/slips\n";
+
+const char report_bw_pps_enhanced_isoch_format[] =
+"[%3d] %4.2f-%4.2f sec  %ss  %ss/sec  %d/%d %8.0f pps  %3d/%d/%d\n";
 
 const char report_sum_bw_pps_enhanced_format[] =
-"[SUM] %4.2f-%4.2f sec  %ss  %ss/sec %4.0f pps\n";
+"[SUM] %4.2f-%4.2f sec  %ss  %ss/sec  %d/%d %8.0f pps\n";
 
 const char report_bw_jitter_loss_enhanced_header[] =
 "[ ID] Interval       Transfer     Bandwidth        Jitter   Lost/Total \
  Latency avg/min/max/stdev PPS\n";
 
 const char report_bw_jitter_loss_enhanced_format[] =
-"[%3d] %4.2f-%4.2f sec  %ss  %ss/sec  %6.3f ms %4"PRIdMAX"/%5"PRIdMAX" (%.2g%%) %6.3f/%6.3f/%6.3f/%6.3f ms %4.0f pps\n";
+"[%3d] %4.2f-%4.2f sec  %ss  %ss/sec  %6.3f ms %4" IPERFdMAX "/%5" IPERFdMAX " (%.2g%%) %6.3f/%6.3f/%6.3f/%6.3f ms %4.0f pps\n";
+
+const char report_bw_jitter_loss_enhanced_isoch_header[] =
+"[ ID] Interval       Transfer     Bandwidth        Jitter   Lost/Total \
+ Latency avg/min/max/stdev PPS Frames/Lost\n";
+
+const char report_bw_jitter_loss_enhanced_isoch_format[] =
+"[%3d] %4.2f-%4.2f sec  %ss  %ss/sec  %6.3f ms %4" IPERFdMAX "/%5" IPERFdMAX " (%.2g%%) %6.3f/%6.3f/%6.3f/%6.3f ms %4.0f pps %3d/%d\n";
 
 const char report_sum_bw_jitter_loss_enhanced_format[] =
-"[SUM] %4.2f-%4.2f sec  %ss  %ss/sec  %6.3f ms %4"PRIdMAX"/%5"PRIdMAX" (%.2g%%)  %4.0f pps\n";
+"[SUM] %4.2f-%4.2f sec  %ss  %ss/sec  %6.3f ms %4" IPERFdMAX "/%5" IPERFdMAX " (%.2g%%)  %4.0f pps\n";
 
 const char report_bw_jitter_loss_suppress_enhanced_format[] =
-"[%3d] %4.2f-%4.2f sec  %ss  %ss/sec  %6.3f ms %4"PRIdMAX"/%5"PRIdMAX" (%.2g%%) -/-/-/- ms %4.0f pps\n";
+"[%3d] %4.2f-%4.2f sec  %ss  %ss/sec  %6.3f ms %4" IPERFdMAX "/%5" IPERFdMAX " (%.2g%%) -/-/-/- ms %4.0f pps\n";
 
 /* -------------------------------------------------------------------
  * Misc reports
@@ -302,6 +338,9 @@ const char report_bw_jitter_loss_suppress_enhanced_format[] =
 const char report_outoforder[] =
 "[%3d] %4.2f-%4.2f sec  %d datagrams received out-of-order\n";
 
+const char report_l2statistics[] =
+"[%3d] %4.2f-%4.2f sec   L2 processing detected errors, total(length/checksum/unknown) = %" IPERFdMAX "(%" IPERFdMAX "/%" IPERFdMAX "/%" IPERFdMAX ")\n";
+
 const char report_sum_outoforder[] =
 "[SUM] %4.2f-%4.2f sec  %d datagrams received out-of-order\n";
 
@@ -326,36 +365,17 @@ const char server_reporting[] =
 const char reportCSV_peer[] =
 "%s,%u,%s,%u";
 
-#ifdef HAVE_QUAD_SUPPORT
-#ifdef HAVE_PRINTF_QD
+const char report_l2length_error[] =
+"[%3d] %4.2f-%4.2f sec  %d datagrams received out-of-order\n";
+
+
 const char reportCSV_bw_format[] =
-"%s,%s,%d,%.1f-%.1f,%qd,%qd\n";
+"%s,%s,%d,%.1f-%.1f,%" IPERFdMAX ",%" IPERFdMAX "\n";
 
 const char reportCSV_bw_jitter_loss_format[] =
-"%s,%s,%d,%.1f-%.1f,%qd,%qd,%.3f,%d,%d,%.3f,%d\n";
-#else // HAVE_PRINTF_QD
-const char reportCSV_bw_format[] =
-"%s,%s,%d,%.1f-%.1f,%lld,%lld\n";
+"%s,%s,%d,%.1f-%.1f,%" IPERFdMAX ",%" IPERFdMAX ",%.3f,%d,%d,%.3f,%d\n";
 
-const char reportCSV_bw_jitter_loss_format[] =
-"%s,%s,%d,%.1f-%.1f,%lld,%lld,%.3f,%d,%d,%.3f,%d\n";
-#endif // HAVE_PRINTF_QD
-#else // HAVE_QUAD_SUPPORT
-#ifdef WIN32
-const char reportCSV_bw_format[] =
-"%s,%s,%d,%.1f-%.1f,%I64d,%I64d\n";
-
-const char reportCSV_bw_jitter_loss_format[] =
-"%s,%s,%d,%.1f-%.1f,%I64d,%I64d,%.3f,%d,%d,%.3f,%d\n";
-#else
-const char reportCSV_bw_format[] =
-"%s,%s,%d,%.1f-%.1f,%d,%d\n";
-
-const char reportCSV_bw_jitter_loss_format[] =
-"%s,%s,%d,%.1f-%.1f,%d,%d,%.3f,%d,%d,%.3f,%d\n";
-#endif //WIN32
-#endif //HAVE_QUAD_SUPPORT
-/* -------------------------------------------------------------------
+ /* -------------------------------------------------------------------
  * warnings
  * ------------------------------------------------------------------- */
 
