@@ -105,27 +105,44 @@ void SetSocketOptions( thread_Settings *inSettings ) {
 #endif
     }
 
-    // check if we're sending multicast, and set TTL
-    if ( isMulticast( inSettings ) && ( inSettings->mTTL > 0 ) ) {
+    // check if we're sending multicast
+    if (isMulticast(inSettings)) {
 #ifdef HAVE_MULTICAST
+	if (!isUDP(inSettings)) {
+	    FAIL(1, "Multicast requires -u option ", inSettings);
+	    exit(1);
+	}
+	// check for default TTL, multicast is 1 and unicast is the system default
+	if (inSettings->mTTL == -1) {
+	    inSettings->mTTL = 1;
+	}
+	if (inSettings->mTTL > 0) {
+	    // set TTL
+	    int val = inSettings->mTTL;
+	    if ( !isIPV6(inSettings) ) {
+		int rc = setsockopt( inSettings->mSock, IPPROTO_IP, IP_MULTICAST_TTL,
+				     (char*) &val, (Socklen_t) sizeof(val));
+
+		WARN_errno( rc == SOCKET_ERROR, "multicast v4 ttl" );
+	    } else
+#  ifdef HAVE_IPV6_MULTICAST
+	    {
+		int rc = setsockopt( inSettings->mSock, IPPROTO_IPV6, IPV6_MULTICAST_HOPS,
+				     (char*) &val, (Socklen_t) sizeof(val));
+		WARN_errno( rc == SOCKET_ERROR, "multicast v6 ttl" );
+	    }
+#  else
+	    FAIL_errno(1, "v6 multicast not supported", inSettings);
+#  endif
+	}
+#endif
+    } else if (inSettings->mTTL > 0) {
 	int val = inSettings->mTTL;
-	if ( !SockAddr_isIPv6( &inSettings->local ) ) {
-	    int rc = setsockopt( inSettings->mSock, IPPROTO_IP, IP_MULTICAST_TTL,
-		    (char*) &val, (Socklen_t) sizeof(val));
-
-	    WARN_errno( rc == SOCKET_ERROR, "multicast ttl" );
-	}
-#ifdef HAVE_IPV6_MULTICAST
-	else {
-	    int rc = setsockopt( inSettings->mSock, IPPROTO_IPV6, IPV6_MULTICAST_HOPS,
-		    (char*) &val, (Socklen_t) sizeof(val));
-	    WARN_errno( rc == SOCKET_ERROR, "multicast ttl" );
-	}
-#endif
-#endif
+	int rc = setsockopt( inSettings->mSock, IPPROTO_IP, IP_TTL,
+			     (char*) &val, (Socklen_t) sizeof(val));
+	WARN_errno( rc == SOCKET_ERROR, "v4 ttl" );
     }
 
-
 #ifdef IP_TOS
 
     // set IP TOS (type-of-service) field
@@ -155,4 +172,22 @@ void SetSocketOptions( thread_Settings *inSettings ) {
 #endif
     }
 }
+
+void SetSocketOptionsSendTimeout( thread_Settings *mSettings, int timer) {
+    if (timer > 0) {
+#ifdef WIN32
+	// Windows SO_SNDTIMEO uses ms
+	DWORD timeout = (double) timer / 1e3;
+#else
+	struct timeval timeout;
+	timeout.tv_sec = timer / 1000000;
+	timeout.tv_usec = timer % 1000000;
+#endif
+	if (setsockopt( mSettings->mSock, SOL_SOCKET, SO_SNDTIMEO, (char *)&timeout, sizeof(timeout)) < 0 ) {
+	    WARN_errno( mSettings->mSock == SO_SNDTIMEO, "socket" );
+	}
+    }
+}
+
+
 // end SetSocketOptions
