@@ -73,9 +73,9 @@ void reporter_printstats( Transfer_Info *stats ) {
     byte_snprintf( buffer, sizeof(buffer)/2, (double) stats->TotalLen,
                    toupper( (int)stats->mFormat));
     byte_snprintf( &buffer[sizeof(buffer)/2], sizeof(buffer)/2,
-                   (stats->TotalLen > 0) ?
-		   stats->TotalLen / (stats->endTime - stats->startTime) : 0,
-                   stats->mFormat);
+                   (!stats->TotalLen || (stats->endTime < SMALLEST_INTERVAL)) ?
+		   0 : (stats->TotalLen / (stats->endTime - stats->startTime)),
+		   stats->mFormat);
     // TCP reports
     if (!stats->mUDP) {
 	if (!stats->mEnhanced) {
@@ -88,50 +88,76 @@ void reporter_printstats( Transfer_Info *stats ) {
 		   buffer, &buffer[sizeof(buffer)/2]);
 	} else {
 	    if( !header_printed ) {
-		printf((stats->mTCP == (char)kMode_Server ? report_bw_read_enhanced_header : report_bw_write_enhanced_header), (stats->tcp.read.binsize/1024.0));
+		printf((stats->mTCP == (char)kMode_Server ? report_bw_read_enhanced_header : report_bw_write_enhanced_header), (stats->sock_callstats.read.binsize/1024.0));
+		header_printed = 1;
 	    }
 	    if (stats->mTCP == (char)kMode_Server) {
 		printf(report_bw_read_enhanced_format,
 		       stats->transferID, stats->startTime, stats->endTime,
 		       buffer, &buffer[sizeof(buffer)/2],
-		       stats->tcp.read.cntRead,
-		       stats->tcp.read.bins[0],
-		       stats->tcp.read.bins[1],
-		       stats->tcp.read.bins[2],
-		       stats->tcp.read.bins[3],
-		       stats->tcp.read.bins[4],
-		       stats->tcp.read.bins[5],
-		       stats->tcp.read.bins[6],
-		       stats->tcp.read.bins[7]);
+		       stats->sock_callstats.read.cntRead,
+		       stats->sock_callstats.read.bins[0],
+		       stats->sock_callstats.read.bins[1],
+		       stats->sock_callstats.read.bins[2],
+		       stats->sock_callstats.read.bins[3],
+		       stats->sock_callstats.read.bins[4],
+		       stats->sock_callstats.read.bins[5],
+		       stats->sock_callstats.read.bins[6],
+		       stats->sock_callstats.read.bins[7]);
 	    } else {
 		printf(report_bw_write_enhanced_format,
 		       stats->transferID, stats->startTime, stats->endTime,
 		       buffer, &buffer[sizeof(buffer)/2],
-		       stats->tcp.write.WriteCnt,
-		       stats->tcp.write.WriteErr,
-		       stats->tcp.write.TCPretry,
-		       stats->tcp.write.cwnd,
-		       stats->tcp.write.rtt);
+		       stats->sock_callstats.write.WriteCnt,
+		       stats->sock_callstats.write.WriteErr,
+		       stats->sock_callstats.write.TCPretry,
+		       stats->sock_callstats.write.cwnd,
+		       stats->sock_callstats.write.rtt);
 	    }
 	}
     } else if ( stats->mUDP == (char)kMode_Client ) {
 	// UDP Client reporting
 	if( !header_printed ) {
+#ifdef HAVE_ISOCHRONOUS
+	    if (stats->mIsochronous)
+		printf( "%s", (stats->mEnhanced ? report_bw_pps_enhanced_isoch_header : report_bw_header));
+	    else
+#endif
 	    printf( "%s", (stats->mEnhanced ? report_bw_pps_enhanced_header : report_bw_header));
 	    header_printed = 1;
 	}
-	printf( stats->mEnhanced ? report_bw_pps_enhanced_format : report_bw_format, stats->transferID,
-		stats->startTime, stats->endTime,
-		buffer, &buffer[sizeof(buffer)/2],
-		(stats->IPGcnt ? (stats->IPGcnt / stats->IPGsum) : 0.0));
+#ifdef HAVE_ISOCHRONOUS
+	if (stats->mIsochronous)
+	    printf( stats->mEnhanced ? report_bw_pps_enhanced_isoch_format : report_bw_format, stats->transferID,
+		    stats->startTime, stats->endTime,
+		    buffer, &buffer[sizeof(buffer)/2],
+		    stats->sock_callstats.write.WriteCnt,
+		    stats->sock_callstats.write.WriteErr,
+		    (stats->IPGcnt ? (stats->IPGcnt / stats->IPGsum) : 0.0),
+		    stats->isochstats.framecnt,
+		    stats->isochstats.framelostcnt, stats->isochstats.slipcnt);
+	else
+#endif
+	    printf( stats->mEnhanced ? report_bw_pps_enhanced_format : report_bw_format, stats->transferID,
+		    stats->startTime, stats->endTime,
+		    buffer, &buffer[sizeof(buffer)/2],
+		    stats->sock_callstats.write.WriteCnt,
+		    stats->sock_callstats.write.WriteErr,
+		    (stats->IPGcnt ? (stats->IPGcnt / stats->IPGsum) : 0.0));
+
     } else {
         // UDP Server Reporting
         if( !header_printed ) {
-	    printf( "%s", (stats->mEnhanced ? report_bw_jitter_loss_enhanced_header : report_bw_jitter_loss_header));
+#ifdef HAVE_ISOCHRONOUS
+	    if (stats->mIsochronous)
+		printf("%s", report_bw_jitter_loss_enhanced_isoch_header);
+	    else
+#endif
+		printf( "%s", (stats->mEnhanced ? report_bw_jitter_loss_enhanced_header : report_bw_jitter_loss_header));
             header_printed = 1;
         }
-	if (stats->IPGcnt) {
-	    if (stats->mEnhanced) {
+	if (stats->mEnhanced) {
+	    if (stats->IPGcnt) {
 		// If the min latency is out of bounds of a realistic value
 		// assume the clocks are not synched and suppress the
 		// latency output
@@ -144,6 +170,21 @@ void reporter_printstats( Transfer_Info *stats ) {
 			    (100.0 * stats->cntError) / stats->cntDatagrams,
 			    (stats->IPGcnt / stats->IPGsum));
 		} else {
+#ifdef HAVE_ISOCHRONOUS
+		    if (stats->mIsochronous) {
+			printf( report_bw_jitter_loss_enhanced_isoch_format, stats->transferID,
+				stats->startTime, stats->endTime,
+				buffer, &buffer[sizeof(buffer)/2],
+				stats->jitter*1000.0, stats->cntError, stats->cntDatagrams,
+				(100.0 * stats->cntError) / stats->cntDatagrams,
+				(stats->transit.sumTransit / stats->transit.cntTransit)*1000.0,
+				stats->transit.minTransit*1000.0,
+				stats->transit.maxTransit*1000.0,
+				(stats->transit.cntTransit < 2) ? 0 : sqrt(stats->transit.m2Transit / (stats->transit.cntTransit - 1)) / 1000,
+				(stats->IPGcnt / stats->IPGsum), stats->isochstats.framecnt, stats->isochstats.framelostcnt);
+		    } else
+#endif
+
 		    printf( report_bw_jitter_loss_enhanced_format, stats->transferID,
 			    stats->startTime, stats->endTime,
 			    buffer, &buffer[sizeof(buffer)/2],
@@ -155,26 +196,53 @@ void reporter_printstats( Transfer_Info *stats ) {
 			    (stats->transit.cntTransit < 2) ? 0 : sqrt(stats->transit.m2Transit / (stats->transit.cntTransit - 1)) / 1000,
 			    (stats->IPGcnt / stats->IPGsum));
 		}
+		if (stats->latency_histogram) {
+		    histogram_print(stats->latency_histogram, stats->startTime, stats->endTime,stats->free);
+		}
+#ifdef HAVE_UDPTRIGGERS
+		if (stats->hostlatency_histogram) {
+		    histogram_print(stats->hostlatency_histogram, stats->startTime, stats->endTime,stats->free);
+		}
+		if (stats->h1_histogram) {
+		    histogram_print(stats->h1_histogram,stats->startTime, stats->endTime,stats->free);
+		    histogram_print(stats->h2_histogram,stats->startTime, stats->endTime,stats->free);
+		    histogram_print(stats->h3_histogram,stats->startTime, stats->endTime,stats->free);
+		    histogram_print(stats->h4_histogram,stats->startTime, stats->endTime,stats->free);
+		    histogram_print(stats->h5_histogram,stats->startTime, stats->endTime,stats->free);
+		    histogram_print(stats->h6_histogram,stats->startTime, stats->endTime,stats->free);
+		}
+#endif
+#ifdef HAVE_ISOCHRONOUS
+		if (stats->framelatency_histogram) {
+		    histogram_print(stats->framelatency_histogram, stats->startTime, stats->endTime,stats->free);
+		}
+#endif
 	    } else {
-		printf( report_bw_jitter_loss_format, stats->transferID,
+		printf( report_bw_jitter_loss_suppress_enhanced_format, stats->transferID,
 			stats->startTime, stats->endTime,
 			buffer, &buffer[sizeof(buffer)/2],
-			stats->jitter*1000.0, stats->cntError, stats->cntDatagrams,
-			(100.0 * stats->cntError) / stats->cntDatagrams);
+			0.0, stats->cntError,
+			stats->cntDatagrams,
+			0.0,0.0,0.0,0.0,0.0,0.0);
 	    }
 	} else {
-	    printf( stats->mEnhanced ? report_bw_jitter_loss_suppress_enhanced_format : report_bw_jitter_loss_format , stats->transferID,
+	    printf( report_bw_jitter_loss_format, stats->transferID,
 		    stats->startTime, stats->endTime,
 		    buffer, &buffer[sizeof(buffer)/2],
-		    0.0, stats->cntError,
-		    stats->cntDatagrams,
-		    0.0,0.0,0.0,0.0,0.0,0.0);
+		    stats->jitter*1000.0, stats->cntError, stats->cntDatagrams,
+		    (100.0 * stats->cntError) / stats->cntDatagrams);
 	}
 	if ( stats->cntOutofOrder > 0 ) {
 	    printf( report_outoforder,
 		    stats->transferID, stats->startTime,
 		    stats->endTime, stats->cntOutofOrder );
 	}
+	if (stats->l2counts.cnt) {
+	    printf( report_l2statistics,
+		    stats->transferID, stats->startTime,
+		    stats->endTime, stats->l2counts.cnt, stats->l2counts.lengtherr,
+		    stats->l2counts.udpcsumerr, stats->l2counts.unknown);
+	}
     }
     // Reset the enhanced stats for the next report interval
     if (stats->mEnhanced) {
@@ -186,11 +254,47 @@ void reporter_printstats( Transfer_Info *stats ) {
 	    stats->transit.vdTransit = 0;
 	    stats->transit.meanTransit = 0;
 	    stats->transit.m2Transit = 0;
+#ifdef HAVE_ISOCHRONOUS
+	    stats->isochstats.framecnt = 0;
+	    stats->isochstats.framelostcnt = 0;
+	    stats->isochstats.slipcnt = 0;
+#endif
 	}
     }
 
-    if ( stats->free == 1 && stats->mUDP == (char)kMode_Client ) {
-        printf( report_datagrams, stats->transferID, stats->cntDatagrams );
+    if ( stats->free == 1) {
+	if (stats->mUDP == (char)kMode_Client ) {
+	    printf( report_datagrams, stats->transferID, stats->cntDatagrams );
+	} else {
+	    if (stats->latency_histogram) {
+		histogram_delete(stats->latency_histogram);
+		stats->latency_histogram = NULL;
+	    }
+#ifdef HAVE_ISOCHRONOUS
+	    if (stats->framelatency_histogram) {
+		histogram_delete(stats->framelatency_histogram);
+		stats->framelatency_histogram = NULL;
+	    }
+#endif
+#ifdef HAVE_UDPTRIGGERS
+	    if (stats->hostlatency_histogram) {
+		histogram_delete(stats->hostlatency_histogram);
+		stats->hostlatency_histogram = NULL;
+	    }
+	    if (stats->h1_histogram) {
+		histogram_delete(stats->h1_histogram);
+		stats->h1_histogram = NULL;
+		histogram_delete(stats->h2_histogram);
+		stats->h2_histogram = NULL;
+		histogram_delete(stats->h3_histogram);
+		stats->h3_histogram = NULL;
+		histogram_delete(stats->h4_histogram);
+		stats->h4_histogram = NULL;
+		histogram_delete(stats->h5_histogram);
+		stats->h5_histogram = NULL;
+	    }
+#endif
+	}
     }
 }
 
@@ -226,6 +330,8 @@ void reporter_multistats( Transfer_Info *stats ) {
 	    printf( report_sum_bw_pps_enhanced_format,
 		    stats->startTime, stats->endTime,
 		    buffer, &buffer[sizeof(buffer)/2],
+		    stats->sock_callstats.write.WriteCnt,
+		    stats->sock_callstats.write.WriteErr,
 		    (stats->IPGcnt ? (stats->IPGcnt / stats->IPGsum) : 0.0));
 	} else {
 	    // TCP Enhanced Reporting
@@ -233,22 +339,22 @@ void reporter_multistats( Transfer_Info *stats ) {
 		printf( report_sum_bw_write_enhanced_format,
 			stats->startTime, stats->endTime,
 			buffer, &buffer[sizeof(buffer)/2],
-			stats->tcp.write.WriteCnt,
-			stats->tcp.write.WriteErr,
-			stats->tcp.write.TCPretry);
+			stats->sock_callstats.write.WriteCnt,
+			stats->sock_callstats.write.WriteErr,
+			stats->sock_callstats.write.TCPretry);
 	    } else {
 		printf( report_sum_bw_read_enhanced_format,
 			stats->startTime, stats->endTime,
 			buffer, &buffer[sizeof(buffer)/2],
-			stats->tcp.read.cntRead,
-			stats->tcp.read.bins[0],
-			stats->tcp.read.bins[1],
-			stats->tcp.read.bins[2],
-			stats->tcp.read.bins[3],
-			stats->tcp.read.bins[4],
-			stats->tcp.read.bins[5],
-			stats->tcp.read.bins[6],
-			stats->tcp.read.bins[7]);
+			stats->sock_callstats.read.cntRead,
+			stats->sock_callstats.read.bins[0],
+			stats->sock_callstats.read.bins[1],
+			stats->sock_callstats.read.bins[2],
+			stats->sock_callstats.read.bins[3],
+			stats->sock_callstats.read.bins[4],
+			stats->sock_callstats.read.bins[5],
+			stats->sock_callstats.read.bins[6],
+			stats->sock_callstats.read.bins[7]);
 	    }
 	}
     }
@@ -296,37 +402,59 @@ void reporter_reportsettings( ReporterData *data ) {
         printf( bind_address, data->mLocalhost );
 	if ((data->mThreadMode != kMode_Client) && \
 	    SockAddr_isMulticast(&data->connection.local)) {
-              printf( join_multicast, data->mLocalhost );
+	    if(!data->mSSMMulticastStr)
+		if (!data->mIfrname)
+		    printf( join_multicast, data->mLocalhost );
+		else
+		    printf(join_multicast_starg_dev, data->mLocalhost,data->mIfrname);
+	    else if(!data->mIfrname)
+		printf( join_multicast_sg, data->mSSMMulticastStr, data->mLocalhost);
+	    else
+		printf( join_multicast_sg_dev, data->mSSMMulticastStr, data->mLocalhost, data->mIfrname);
         }
     }
 
-    if ( isUDP( data ) ) {
-
-	if (data->mThreadMode != kMode_Listener) {
-	    double delay_target;
-	    if (data->mUDPRateUnits == kRate_BW) {
-		delay_target = (double) ( data->mBufLen * 8000000.0 / data->mUDPRate);
-	    } else {
-		delay_target = (1e6 / data->mUDPRate);
-	    }
-#ifdef HAVE_KALMAN
-	    printf(client_datagram_size_kalman, data->mBufLen, delay_target);
+    if (isIsochronous(data)) {
+#ifdef HAVE_ISOCHRONOUS
+	char meanbuf[40];
+	char variancebuf[40];
+	byte_snprintf(meanbuf, sizeof(meanbuf), data->isochstats.mMean, 'a');
+	byte_snprintf(variancebuf, sizeof(variancebuf), data->isochstats.mVariance, 'a');
+	printf(client_udp_isochronous, data->isochstats.mFPS, meanbuf, variancebuf, (data->isochstats.mBurstInterval/1000.0), (data->isochstats.mBurstIPG/1000.0));
 #else
-	    printf(client_datagram_size, data->mBufLen, delay_target);
+	fprintf(stderr, "--isochronous not supportted, try --enable-isochronous during config and remake\n");
 #endif
-	} else {
-	    printf(server_datagram_size, data->mBufLen);
+    } else {
+	if ( isUDP( data ) ) {
+	    if (data->mThreadMode != kMode_Listener) {
+		double delay_target;
+		if (data->mUDPRateUnits == kRate_BW) {
+		    delay_target = (double) ( data->mBufLen * 8000000.0 / data->mUDPRate);
+		} else {
+		    delay_target = (1e6 / data->mUDPRate);
+		}
+		if (isTxSync(data)) {
+		    printf(client_datagram_size_tx_sync, data->mBufLen, (data->TxSyncInterval * 1000.0));
+		} else {
+#ifdef HAVE_KALMAN
+		    printf(client_datagram_size_kalman, data->mBufLen, delay_target);
+#else
+		    printf(client_datagram_size, data->mBufLen, delay_target);
+#endif
+		}
+	    } else {
+		printf(server_datagram_size, data->mBufLen);
+	    }
+	    if ( SockAddr_isMulticast( &data->connection.peer ) ) {
+		printf( multicast_ttl, data->info.mTTL);
+	    }
+	} else if (isEnhanced(data)) {
+	    byte_snprintf( buffer, sizeof(buffer), data->mBufLen,
+			   toupper( (int)data->info.mFormat));
+	    printf("%s: %s\n", ((data->mThreadMode == kMode_Client) ?
+				client_write_size : server_read_size), buffer);
 	}
-        if ( SockAddr_isMulticast( &data->connection.peer ) ) {
-            printf( multicast_ttl, data->info.mTTL);
-        }
-    } else if (isEnhanced(data)) {
-	byte_snprintf( buffer, sizeof(buffer), data->mBufLen,
-		       toupper( (int)data->info.mFormat));
-	printf("%s: %s\n", ((data->mThreadMode == kMode_Client) ?
-			   client_write_size : server_read_size), buffer);
     }
-
     byte_snprintf( buffer, sizeof(buffer), win,
                    toupper( (int)data->info.mFormat));
     printf( "%s: %s", (isUDP( data ) ?
@@ -353,6 +481,17 @@ void *reporter_reportpeer( Connection_Info *stats, int ID ) {
         struct sockaddr *local = ((struct sockaddr*)&stats->local);
         struct sockaddr *peer = ((struct sockaddr*)&stats->peer);
 
+	char extbuf[2*PEERBUFSIZE];
+	char *b = &extbuf[0];
+	extbuf[0]= '\0';
+	if (stats->l2mode) {
+	    snprintf(b, PEERBUFSIZE, " (%s=0x%X)", "l2mode", stats->l2mode);
+	    b += strlen(b);
+	}
+	if (stats->peerversion) {
+	    snprintf(b, PEERBUFSIZE, "%s", stats->peerversion);
+	}
+
         if ( local->sa_family == AF_INET ) {
             inet_ntop( AF_INET, &((struct sockaddr_in*)local)->sin_addr,
                        local_addr, REPORT_ADDRLEN);
@@ -374,8 +513,7 @@ void *reporter_reportpeer( Connection_Info *stats, int ID ) {
                        remote_addr, REPORT_ADDRLEN);
         }
 #endif
-
-        printf( report_peer,
+	 printf( report_peer,
                 ID,
                 local_addr,  ( local->sa_family == AF_INET ?
                               ntohs(((struct sockaddr_in*)local)->sin_port) :
@@ -389,10 +527,9 @@ void *reporter_reportpeer( Connection_Info *stats, int ID ) {
 #ifdef HAVE_IPV6
 			      ntohs(((struct sockaddr_in6*)peer)->sin6_port)),
 #else
-                              0),
-
+	                      0),
 #endif
-	         stats->peerversion);
+	         extbuf);
     }
     return NULL;
 }
@@ -406,7 +543,11 @@ void reporter_peerversion (thread_Settings *inSettings, int upper, int lower) {
     major = (upper & 0x0000FFFF);
     minor = (lower & 0xFFFF0000) >> 16;
     alpha = (lower & 0x0000000F);
-    sprintf(inSettings->peerversion," (peer %d.%d.%d)", rel, major, minor);
+    if (!isIsochronous(inSettings)) {
+	sprintf(inSettings->peerversion," (peer %d.%d.%d)", rel, major, minor);
+    } else {
+	sprintf(inSettings->peerversion," isoch (peer %d.%d.%d)", rel, major, minor);
+    }
     switch(alpha) {
     case 0:
 	sprintf(inSettings->peerversion + strlen(inSettings->peerversion) - 1,"-alpha)");
@@ -422,6 +563,10 @@ void reporter_peerversion (thread_Settings *inSettings, int upper, int lower) {
     default:
 	sprintf(inSettings->peerversion + strlen(inSettings->peerversion) - 1, "-unk)");
     }
+    if (isUDPTriggers(inSettings)) {
+	int len=strlen(inSettings->peerversion);
+	snprintf(&inSettings->peerversion[len], PEERBUFSIZE, " (%s)", "triggers");
+    }
 }
 /* -------------------------------------------------------------------
  * Report the MSS and MTU, given the MSS (or a guess thereof)
