@@ -58,6 +58,7 @@
 #include "Locale.h"
 #include "PerfSocket.hpp"
 #include "SocketAddr.h"
+#include "histogram.h"
 
 #ifdef __cplusplus
 extern "C" {
@@ -109,7 +110,7 @@ report_statistics multiple_reports[kReport_MAXIMUM] = {
     CSV_stats
 };
 
-char buffer[64]; // Buffer for printing
+char buffer[SNBUFFERSIZE]; // Buffer for printing
 ReportHeader *ReportRoot = NULL;
 static int num_multi_slots = 0;
 extern Condition ReportCond;
@@ -134,10 +135,10 @@ MultiHeader* InitMulti( thread_Settings *agent, int inID) {
 		num_multi_slots = (((agent->mThreads * 2) + 1) > NUM_MULTI_SLOTS) ? ((agent->mThreads * 2) + 1) : NUM_MULTI_SLOTS;
 	    }
 	    // printf ("Alloc %d multislots\n", num_multi_slots);
-            multihdr = malloc(sizeof(MultiHeader) +  sizeof(ReporterData) +
-                              num_multi_slots * sizeof(Transfer_Info));
+            multihdr = calloc((sizeof(MultiHeader) +  sizeof(ReporterData) +
+			       num_multi_slots * sizeof(Transfer_Info)), sizeof(char*));
         } else {
-            multihdr = malloc(sizeof(MultiHeader));
+            multihdr = calloc(sizeof(MultiHeader), sizeof(char*));
         }
         if ( multihdr != NULL ) {
             memset( multihdr, 0, sizeof(MultiHeader) );
@@ -174,7 +175,7 @@ MultiHeader* InitMulti( thread_Settings *agent, int inID) {
                 data->info.mFormat = agent->mFormat;
                 data->info.mTTL = agent->mTTL;
 		if (data->mThreadMode == kMode_Server)
-		    data->info.tcp.read.binsize = data->mBufLen / 8;
+		    data->info.sock_callstats.read.binsize = data->mBufLen / 8;
                 if ( isEnhanced( agent ) ) {
 		    data->info.mEnhanced = 1;
 		} else {
@@ -236,11 +237,9 @@ ReportHeader* InitReport( thread_Settings *agent ) {
         /*
          * Create in one big chunk
          */
-        reporthdr = malloc( sizeof(ReportHeader) +
-                            NUM_REPORT_STRUCTS * sizeof(ReportStruct) );
+        reporthdr = calloc( sizeof(ReportHeader) +
+                            NUM_REPORT_STRUCTS * sizeof(ReportStruct), sizeof(char*));
         if ( reporthdr != NULL ) {
-            // Only need to make sure the headers are clean
-            memset( reporthdr, 0, sizeof(ReportHeader));
             reporthdr->data = (ReportStruct*)(reporthdr+1);
             reporthdr->multireport = agent->multihdr;
             data = &reporthdr->report;
@@ -255,16 +254,19 @@ ReportHeader* InitReport( thread_Settings *agent ) {
             }
             data->mHost = agent->mHost;
             data->mLocalhost = agent->mLocalhost;
+	    data->mSSMMulticastStr = agent->mSSMMulticastStr;
+	    data->mIfrname = agent->mIfrname;
             data->mBufLen = agent->mBufLen;
             data->mMSS = agent->mMSS;
             data->mTCPWin = agent->mTCPWin;
             data->flags = agent->flags;
             data->mThreadMode = agent->mThreadMode;
             data->mode = agent->mReportMode;
+	    data->TxSyncInterval = agent->mTxSyncInterval;
             data->info.mFormat = agent->mFormat;
             data->info.mTTL = agent->mTTL;
 	    if (data->mThreadMode == kMode_Server)
-		data->info.tcp.read.binsize = data->mBufLen / 8;
+		data->info.sock_callstats.read.binsize = data->mBufLen / 8;
             if ( isUDP( agent ) ) {
 		gettimeofday(&data->IPGstart, NULL);
                 reporthdr->report.info.mUDP = (char)agent->mThreadMode;
@@ -276,6 +278,74 @@ ReportHeader* InitReport( thread_Settings *agent ) {
 	    } else {
 		data->info.mEnhanced = 0;
 	    }
+	    if (data->mThreadMode == kMode_Server) {
+		if (isUDPHistogram(agent)) {
+		    char name[] = "T8";
+		    data->info.latency_histogram =  histogram_init(agent->mUDPbins,agent->mUDPbinsize,0,\
+								   (agent->mUDPunits ? 1e6 : 1e3), \
+								   agent->mUDPci_lower, agent->mUDPci_upper, data->info.transferID, name);
+		}
+#ifdef HAVE_UDPTRIGGERS
+                /*
+		 * TSF histograms
+		 * DHDRX-DHDTX 'DRRx-DRTx;'
+		 * hs1 = 14,8 T6 "FWR2-FWT1"
+		 * hs2 = 15,7 T5 "FWR1-FWT2'
+		 * hs3 = 14,17 T4 'FWT4-FWT1'
+		 * hs4 = 15,16 T3 'FWT3-FWT2'
+		 * hs5 = 7,8 T2 'FWR2-FWR1'
+		 * hs6 = 15, 14  "FWT2-FWT1'
+		 */
+		if (isUDPTriggers(agent)) {
+		    char name[] = "DRRx-DRTx";
+		    // Bins, bin size, bin, offset, bin units (1e6 = us), confidence low (e.g. 5%), confidence high (e.g. 95%), id, name)
+		    data->info.hostlatency_histogram =  histogram_init(100000,10, 0, 1e6, 5, 95,  \
+								       data->info.transferID, name);
+		    strcpy(name,"FWR2-FWT1");
+		    data->info.h1_histogram =  histogram_init(1000000,10, 0, 1e6, 5, 95, data->info.transferID, name);
+		    strcpy(name,"FWR1-FWT2");
+		    data->info.h2_histogram =  histogram_init(1000000,10, 0, 1e6, 5, 95, data->info.transferID, name);
+		    strcpy(name,"FWT4-FWT1");
+		    data->info.h3_histogram =  histogram_init(1000000,10, 0, 1e6, 5, 95, data->info.transferID, name);
+		    strcpy(name,"FWT3-FWT2");
+		    data->info.h4_histogram =  histogram_init(1000000,10, 0, 1e6, 5, 95, data->info.transferID, name);
+		    strcpy(name,"FWR2-FWR1");
+		    data->info.h5_histogram =  histogram_init(1000000,10, 0, 1e6, 5, 95, data->info.transferID, name);
+		    strcpy(name,"FWT2-FWT1");
+		    data->info.h6_histogram =  histogram_init(1000000,10, 0, 1e6, 5, 95, data->info.transferID, name);
+
+		    memset(&data->info.tsftv_rxmac,0,sizeof(struct tsftv_t));
+//		    memcpy(&data->info.tsftv_rxpcie, &data->info.tsftv_rxmac, sizeof(struct tsftv_t));
+		    memcpy(&data->info.tsftv_txpcie, &data->info.tsftv_rxmac, sizeof(struct tsftv_t));
+		    memcpy(&data->info.tsftv_txpciert, &data->info.tsftv_rxmac, sizeof(struct tsftv_t));
+		    memcpy(&data->info.tsftv_txdma, &data->info.tsftv_rxmac, sizeof(struct tsftv_t));
+		    memcpy(&data->info.tsftv_txstatus, &data->info.tsftv_rxmac, sizeof(struct tsftv_t));
+
+		    // sync the RX tsf structs with gps
+		    tsfgps_sync(&data->info.tsftv_rxmac, NULL, agent);
+		    memcpy(&data->info.tsftv_rxpcie, &data->info.tsftv_rxmac, sizeof(struct tsftv_t));
+		}
+#endif
+#ifdef HAVE_ISOCHRONOUS
+		if (isUDPHistogram(agent) && isIsochronous(agent)) {
+		    char name[] = "F8";
+		    // make sure frame bin size min is 100 microsecond
+		    if (agent->mUDPunits && (agent->mUDPbinsize < 100))
+			agent->mUDPbinsize = 100;
+		    agent->mUDPunits = 1;
+		    data->info.framelatency_histogram =  histogram_init(agent->mUDPbins,agent->mUDPbinsize,0, \
+									(agent->mUDPunits ? 1e6 : 1e3), agent->mUDPci_lower, \
+									agent->mUDPci_upper, data->info.transferID, name);
+		}
+#endif
+	    }
+#ifdef HAVE_ISOCHRONOUS
+	    if ( isIsochronous( agent ) ) {
+		data->info.mIsochronous = 1;
+	    } else {
+		data->info.mIsochronous = 0;
+	    }
+#endif
         } else {
             FAIL(1, "Out of Memory!!\n", agent);
         }
@@ -285,10 +355,8 @@ ReportHeader* InitReport( thread_Settings *agent ) {
             /*
              * Create in one big chunk
              */
-            reporthdr = malloc( sizeof(ReportHeader) );
+            reporthdr = calloc( sizeof(ReportHeader), sizeof(char*) );
             if ( reporthdr != NULL ) {
-                // Only need to make sure the headers are clean
-                memset( reporthdr, 0, sizeof(ReportHeader));
                 data = &reporthdr->report;
                 data->info.transferID = agent->mSock;
                 data->info.groupID = -1;
@@ -303,6 +371,10 @@ ReportHeader* InitReport( thread_Settings *agent ) {
             data->connection.local = agent->local;
             data->connection.size_local = agent->size_local;
 	    data->connection.peerversion = agent->peerversion;
+	    // Set the l2mode flags
+	    data->connection.l2mode = isL2LengthCheck(agent);
+	    if (data->connection.l2mode)
+		data->connection.l2mode = ((isIPV6(agent) << 1) | data->connection.l2mode);
         } else {
             FAIL(1, "Out of Memory!!\n", agent);
         }
@@ -467,8 +539,7 @@ void ReportSettings( thread_Settings *agent ) {
         /*
          * Create in one big chunk
          */
-        ReportHeader *reporthdr = malloc( sizeof(ReportHeader) );
-
+        ReportHeader *reporthdr = calloc( sizeof(ReportHeader), sizeof(char*));
         if ( reporthdr != NULL ) {
             ReporterData *data = &reporthdr->report;
             data->info.transferID = agent->mSock;
@@ -478,12 +549,15 @@ void ReportSettings( thread_Settings *agent ) {
 
             data->mHost = agent->mHost;
             data->mLocalhost = agent->mLocalhost;
+	    data->mSSMMulticastStr = agent->mSSMMulticastStr;
+	    data->mIfrname = agent->mIfrname;
             data->mode = agent->mReportMode;
             data->type = SETTINGS_REPORT;
             data->mBufLen = agent->mBufLen;
             data->mMSS = agent->mMSS;
             data->mTCPWin = agent->mTCPWin;
             data->flags = agent->flags;
+            data->flags_extend = agent->flags_extend;
             data->mThreadMode = agent->mThreadMode;
             data->mPort = agent->mPort;
             data->info.mFormat = agent->mFormat;
@@ -494,6 +568,16 @@ void ReportSettings( thread_Settings *agent ) {
             data->connection.size_local = agent->size_local;
             data->mUDPRate = agent->mUDPRate;
             data->mUDPRateUnits = agent->mUDPRateUnits;
+	    data->TxSyncInterval = agent->mTxSyncInterval;
+#ifdef HAVE_ISOCHRONOUS
+	    if (isIsochronous(data)) {
+		data->isochstats.mFPS = agent->mFPS;
+		data->isochstats.mMean = agent->mMean/8;
+		data->isochstats.mVariance = agent->mVariance/8;
+		data->isochstats.mBurstIPG = (unsigned int) (agent->mBurstIPG*1000.0);
+		data->isochstats.mBurstInterval = (unsigned int) (1 / agent->mFPS * 1000000);
+	    }
+#endif
     #ifdef HAVE_THREAD
             /*
              * Update the ReportRoot to include this report.
@@ -528,7 +612,7 @@ void ReportServerUDP( thread_Settings *agent, server_hdr *server ) {
 	/*
 	 * Create in one big chunk
 	 */
-	ReportHeader *reporthdr = malloc( sizeof(ReportHeader) );
+	ReportHeader *reporthdr = calloc( sizeof(ReportHeader), sizeof(char*));
 	Transfer_Info *stats = &reporthdr->report.info;
 
 	if ( !reporthdr ) {
@@ -546,8 +630,12 @@ void ReportServerUDP( thread_Settings *agent, server_hdr *server ) {
 	stats->mFormat = agent->mFormat;
 	stats->jitter = ntohl( server->base.jitter1 );
 	stats->jitter += ntohl( server->base.jitter2 ) / (double)rMillion;
+#ifdef HAVE_QUAD_SUPPORT
 	stats->TotalLen = (((max_size_t) ntohl( server->base.total_len1 )) << 32) + \
 	    ntohl( server->base.total_len2 );
+#else
+	stats->TotalLen = (max_size_t) ntohl(server->base.total_len2);
+#endif
 	stats->startTime = 0;
 	stats->endTime = ntohl( server->base.stop_sec );
 	stats->endTime += ntohl( server->base.stop_usec ) / (double)rMillion;
@@ -556,8 +644,12 @@ void ReportServerUDP( thread_Settings *agent, server_hdr *server ) {
 #ifndef HAVE_SEQNO64b
 	stats->cntDatagrams = ntohl( server->base.datagrams );
 #else
+  #ifdef HAVE_QUAD_SUPPORT
 	stats->cntDatagrams = (((max_size_t) ntohl( server->base.datagrams2 )) << 32) + \
 	    ntohl( server->base.datagrams );
+  #else
+        stats->TotalLen = (max_size_t) ntohl(server->base.datagrams);
+  #endif
 #endif
 	if ((flags & HEADER_EXTEND) != 0) {
 	    stats->mEnhanced = 1;
@@ -641,6 +733,33 @@ again:
                     itr->next = temp->next;
                 }
                 // finished with report so free it
+		if (temp->report.info.latency_histogram) {
+		    histogram_delete(temp->report.info.latency_histogram);
+		}
+#ifdef HAVE_ISOCHRONOUS
+		if (temp->report.info.framelatency_histogram) {
+		    histogram_delete(temp->report.info.framelatency_histogram);
+		}
+#endif
+#ifdef HAVE_UDPTRIGGERS
+		if (temp->report.info.hostlatency_histogram) {
+		    histogram_delete(temp->report.info.hostlatency_histogram);
+		}
+		if (temp->report.info.h1_histogram) {
+		    histogram_delete(temp->report.info.h1_histogram);
+		    temp->report.info.h1_histogram = NULL;
+		    histogram_delete(temp->report.info.h2_histogram);
+		    temp->report.info.h2_histogram = NULL;
+		    histogram_delete(temp->report.info.h3_histogram);
+		    temp->report.info.h3_histogram = NULL;
+		    histogram_delete(temp->report.info.h4_histogram);
+		    temp->report.info.h4_histogram = NULL;
+		    histogram_delete(temp->report.info.h5_histogram);
+		    temp->report.info.h5_histogram = NULL;
+		    histogram_delete(temp->report.info.h6_histogram);
+		    temp->report.info.h6_histogram = NULL;
+		}
+#endif
                 free( temp );
                 Condition_Unlock ( ReportCond );
                 Condition_Signal( &ReportDoneCond );
@@ -703,6 +822,19 @@ again:
 void process_report ( ReportHeader *report ) {
     if ( report != NULL ) {
         if ( reporter_process_report( report ) ) {
+	    if (report->report.info.latency_histogram) {
+		histogram_delete(report->report.info.latency_histogram);
+	    }
+#ifdef HAVE_ISOCHRONOUS
+	    if (report->report.info.framelatency_histogram) {
+		histogram_delete(report->report.info.framelatency_histogram);
+	    }
+#endif
+#ifdef HAVE_UDPTRIGGERS
+	    if (report->report.info.hostlatency_histogram) {
+		histogram_delete(report->report.info.hostlatency_histogram);
+	    }
+#endif
             free( report );
         }
     }
@@ -720,6 +852,19 @@ int reporter_process_report ( ReportHeader *reporthdr ) {
             // If we are done with this report then free it
             ReportHeader *temp = reporthdr->next;
             reporthdr->next = reporthdr->next->next;
+	    if (temp->report.info.latency_histogram) {
+		histogram_delete(temp->report.info.latency_histogram);
+	    }
+#ifdef HAVE_ISOCHRONOUS
+	    if (temp->report.info.framelatency_histogram) {
+		histogram_delete(temp->report.info.framelatency_histogram);
+	    }
+#endif
+#ifdef HAVE_UDPTRIGGERS
+	    if (temp->report.info.hostlatency_histogram) {
+		histogram_delete(temp->report.info.hostlatency_histogram);
+	    }
+#endif
             free( temp );
         }
     }
@@ -774,6 +919,7 @@ int reporter_process_report ( ReportHeader *reporthdr ) {
 /*
  * Updates connection stats
  */
+#define L2DROPFILTERCOUNTER 100
 int reporter_handle_packet( ReportHeader *reporthdr ) {
     ReportStruct *packet = &reporthdr->data[reporthdr->reporterindex];
     ReporterData *data = &reporthdr->report;
@@ -789,21 +935,181 @@ int reporter_handle_packet( ReportHeader *reporthdr ) {
             data->TotalLen += packet->packetLen;
         }
     } else {
+	// Process error counters that are mostly
+	// unrelated to receiving a valid packet
+	//
+	// First, are client socket write counters
+	if (reporthdr->report.mThreadMode == kMode_Client) {
+	    if (packet->errwrite) {
+		stats->sock_callstats.write.WriteErr++;
+		stats->sock_callstats.write.totWriteErr++;
+	    } else {
+		stats->sock_callstats.write.WriteCnt++;
+		stats->sock_callstats.write.totWriteCnt++;
+	    }
+	// Next are server l2 errors, filter out first n L2 errors
+	// due to BPF AF_PACKET race
+	} else if (packet->l2errors && (data->cntDatagrams > L2DROPFILTERCOUNTER)) {
+	    stats->l2counts.cnt++;
+	    stats->l2counts.tot_cnt++;
+	    if (packet->l2errors & L2UNKNOWN) {
+		stats->l2counts.unknown++;
+		stats->l2counts.tot_unknown++;
+	    }
+	    if (packet->l2errors & L2LENERR) {
+		stats->l2counts.lengtherr++;
+		stats->l2counts.tot_lengtherr++;
+	    }
+	    if (packet->l2errors & L2CSUMERR) {
+		stats->l2counts.udpcsumerr++;
+		stats->l2counts.tot_udpcsumerr++;
+	    }
+	}
+	// These are valid packets that need standard iperf accounting
 	if (!packet->emptyreport) {
 	    // update fields common to TCP and UDP, client and server
 	    data->TotalLen += packet->packetLen;
+	    // update fields common to TCP and UDP client
 	    // update fields common to UDP client and server
             if ( isUDP( data ) ) {
 		data->cntDatagrams++;
 		stats->IPGsum += TimeDifference(data->packetTime, data->IPGstart );
 		stats->IPGcnt++;
 		data->IPGstart = data->packetTime;
+#ifdef HAVE_ISOCHRONOUS
+		{
+		    int framedelta=0;
+		    // very first isochronous frame
+		    if (!data->isochstats.frameID) {
+			data->isochstats.framecnt=packet->frameID;
+			data->isochstats.framecnt=1;
+			stats->isochstats.framecnt=1;
+		    } else {
+			static int matchframeid=0;
+			// perform client and server frame based accounting
+			framedelta = (packet->frameID - data->isochstats.frameID);
+			if (framedelta) {
+			    data->isochstats.framecnt++;
+			    stats->isochstats.framecnt++;
+			    if (framedelta > 1) {
+				if (stats->mUDP == kMode_Server) {
+				    int lost = framedelta - (packet->frameID - packet->prevframeID);
+				    stats->isochstats.framelostcnt += lost;
+				    data->isochstats.framelostcnt += lost;
+				} else {
+				    stats->isochstats.framelostcnt += (framedelta-1);
+				    data->isochstats.framelostcnt += (framedelta-1);
+				    stats->isochstats.slipcnt++;
+				    data->isochstats.slipcnt++;
+				}
+			    }
+			}
+			// peform frame latency checks
+			if (stats->framelatency_histogram) {
+			    // first packet of a burst and not a duplicate
+			    if ((packet->burstsize == packet->remaining) && (matchframeid!=packet->frameID)) {
+				matchframeid=packet->frameID;
+			    }
+			    if ((packet->packetLen == packet->remaining) && (packet->frameID == matchframeid)) {
+				// last packet of a burst (or first-last in case of a duplicate) and frame id match
+				double frametransit = TimeDifference(packet->packetTime, packet->isochStartTime) \
+				    - ((packet->burstperiod * (packet->frameID - 1)) / 1000000.0);
+			        histogram_insert(stats->framelatency_histogram, frametransit);
+				matchframeid = 0;  // reset the matchid so any potential duplicate is ignored
+			    }
+			}
+		    }
+		    data->isochstats.frameID = packet->frameID;
+		}
+#endif
 		// Finally, update UDP server fields
 		if (stats->mUDP == kMode_Server) {
 		    //subsequent packets
 		    double transit;
 		    double deltaTransit;
 		    transit = TimeDifference( packet->packetTime, packet->sentTime );
+		    if (stats->latency_histogram) {
+			histogram_insert(stats->latency_histogram, transit);
+		    }
+#ifdef HAVE_UDPTRIGGERS
+		    if (stats->hostlatency_histogram) {
+			double hosttransit = TimeDifference(packet->hostRxTime, packet->hostTxTime);
+			histogram_insert(stats->hostlatency_histogram, hosttransit);
+		    }
+
+		    if (stats->h1_histogram) {
+			// apply sync to the TX tsf structs if needed
+			if (!stats->tsftv_txpcie.synced) {
+			    struct gpsref_sync_t tsfgps;
+			    tsfgps.gps_ts.tv_sec = packet->gps_sync.tv_sec;
+			    tsfgps.gps_ts.tv_nsec = packet->gps_sync.tv_nsec;
+			    tsfgps.ref_ts.tv_sec = packet->ref_sync.tv_sec;
+			    tsfgps.ref_ts.tv_nsec = packet->ref_sync.tv_nsec;
+			    tsfgps_sync(&stats->tsftv_txpcie, &tsfgps, NULL);
+			    tsfgps_sync(&stats->tsftv_txpciert, &tsfgps, NULL);
+			    tsfgps_sync(&stats->tsftv_txdma, &tsfgps, NULL);
+			    tsfgps_sync(&stats->tsftv_txstatus, &tsfgps, NULL);
+			}
+			int ix;
+			for (ix = 0; ix < packet->tsfcount; ix ++) {
+			    /*
+			     * TSF histograms
+			     * hs1 = 14,8
+			     * hs2 = 15,7
+			     * hs3 = 14,17
+			     * hs4 = 15,16
+			     * hs5 = 7,8 (7-1,8-2)
+			     * hs6 = 14,15
+			     *
+			     * u_int32_t tsf_rxmac   7
+			     * u_int32_t tsf_rxpcie  8
+			     * u_int32_t tsf_txpcie  14
+			     * u_int32_t tsf_txdma   15
+			     * u_int32_t tsf_txstatus 16
+			     * u_int32_t tsf_txpciert  17
+			     */
+			    // A TSF of all ones indicates invalid, ignore those samples
+			    if ((packet->tsf[ix].tsf_rxmac != 0xFFFFFFFF) && \
+				(packet->tsf[ix].tsf_rxpcie != 0xFFFFFFFF) && \
+				(packet->tsf[ix].tsf_txpcie != 0xFFFFFFFF) && \
+				(packet->tsf[ix].tsf_txdma != 0xFFFFFFFF) && \
+				(packet->tsf[ix].tsf_txstatus != 0xFFFFFFFF) && \
+				(packet->tsf[ix].tsf_txpciert != 0xFFFFFFFF)) {
+				// Update the TSF timestamp structures using the raw timestamps
+				tsfraw_update(&stats->tsftv_rxpcie, packet->tsf[ix].tsf_rxpcie);
+				tsfraw_update(&stats->tsftv_rxmac, packet->tsf[ix].tsf_rxmac);
+				tsfraw_update(&stats->tsftv_txpcie, packet->tsf[ix].tsf_txpcie);
+				tsfraw_update(&stats->tsftv_txpciert, packet->tsf[ix].tsf_txpciert);
+				tsfraw_update(&stats->tsftv_txdma, packet->tsf[ix].tsf_txdma);
+				tsfraw_update(&stats->tsftv_txstatus, packet->tsf[ix].tsf_txstatus);
+				// compute the diffs
+				float hs1 = tsf_sec_delta(&stats->tsftv_txpcie, &stats->tsftv_rxpcie);
+				float hs2 = tsf_sec_delta(&stats->tsftv_txdma, &stats->tsftv_rxmac);
+				float hs3 = tsf_sec_delta(&stats->tsftv_txpcie, &stats->tsftv_txpciert);
+				float hs4 = tsf_sec_delta(&stats->tsftv_txdma, &stats->tsftv_txstatus);
+				float hs5 = tsf_sec_delta(&stats->tsftv_rxmac, &stats->tsftv_rxpcie);
+				float hs6 = tsf_sec_delta(&stats->tsftv_txpcie, &stats->tsftv_txdma);
+#if 0
+				{
+				    fprintf(stderr, "TSF Debug: hs1=%f\n", hs1);
+				    fprintf(stderr, "TSF Debug: hs2=%f\n", hs2);
+				    fprintf(stderr, "TSF Debug: hs3=%f\n", hs3);
+				    fprintf(stderr, "TSF Debug: hs4=%f\n", hs4);
+				    fprintf(stderr, "TSF Debug: hs5=%f\n", hs5);
+				    fprintf(stderr, "TSF Debug: hs6=%f\n", hs6);
+				}
+#endif
+				histogram_insert(stats->h1_histogram, hs1);
+				histogram_insert(stats->h2_histogram, hs2);
+				histogram_insert(stats->h3_histogram, hs3);
+				histogram_insert(stats->h4_histogram, hs4);
+				histogram_insert(stats->h5_histogram, hs5);
+				histogram_insert(stats->h6_histogram, hs6);
+			    }
+			}
+		    }
+#endif
+
 		    // packet loss occured if the datagram numbers aren't sequential
 		    if ( packet->packetID != data->PacketID + 1 ) {
 			if (packet->packetID < data->PacketID + 1 ) {
@@ -877,20 +1183,11 @@ int reporter_handle_packet( ReportHeader *reporthdr ) {
 	    } else if (reporthdr->report.mThreadMode == kMode_Server && (packet->packetLen > 0)) {
 		int bin;
 		// mean min max tests
-		stats->tcp.read.cntRead++;
-		stats->tcp.read.totcntRead++;
-		bin = (int)floor((packet->packetLen -1)/stats->tcp.read.binsize);
-		stats->tcp.read.bins[bin]++;
-		stats->tcp.read.totbins[bin]++;
-	    } else if (reporthdr->report.mThreadMode == kMode_Client) {
-		if (packet->errwrite) {
-		    stats->tcp.write.WriteErr++;
-		    stats->tcp.write.totWriteErr++;
-		}
-		else {
-		    stats->tcp.write.WriteCnt++;
-		    stats->tcp.write.totWriteCnt++;
-		}
+		stats->sock_callstats.read.cntRead++;
+		stats->sock_callstats.read.totcntRead++;
+		bin = (int)floor((packet->packetLen -1)/stats->sock_callstats.read.binsize);
+		stats->sock_callstats.read.bins[bin]++;
+		stats->sock_callstats.read.totbins[bin]++;
 	    }
 	} else if ((stats->mUDP == kMode_Server) &&	\
 		   (stats->transit.cntTransit == 0)) {
@@ -949,15 +1246,17 @@ void reporter_handle_multiple_reports( MultiHeader *reporthdr, Transfer_Info *st
 		current->mTCP = stats->mTCP;
 		if (stats->mTCP == kMode_Server) {
 		    int ix;
-		    current->tcp.read.cntRead = stats->tcp.read.cntRead;
+		    current->sock_callstats.read.cntRead = stats->sock_callstats.read.cntRead;
 		    for (ix = 0; ix < 8; ix++) {
-			current->tcp.read.bins[ix] = stats->tcp.read.bins[ix];
+			current->sock_callstats.read.bins[ix] = stats->sock_callstats.read.bins[ix];
 		    }
-		} else if (stats->mTCP == kMode_Client) {
-		    current->tcp.write.WriteErr = stats->tcp.write.WriteErr;
-		    current->tcp.write.WriteCnt = stats->tcp.write.WriteCnt;
+		} else {
+		    current->sock_callstats.write.WriteErr = stats->sock_callstats.write.WriteErr;
+		    current->sock_callstats.write.WriteCnt = stats->sock_callstats.write.WriteCnt;
 #ifdef HAVE_STRUCT_TCP_INFO_TCPI_TOTAL_RETRANS
-		    current->tcp.write.TCPretry = stats->tcp.write.TCPretry;
+		    if (stats->mTCP == kMode_Client) {
+			current->sock_callstats.write.TCPretry = stats->sock_callstats.write.TCPretry;
+		    }
 #endif
 		}
                 current->free = 1;
@@ -969,15 +1268,17 @@ void reporter_handle_multiple_reports( MultiHeader *reporthdr, Transfer_Info *st
 		current->IPGcnt += stats->IPGcnt;
 		if (stats->mTCP == kMode_Server) {
 		    int ix;
-		    current->tcp.read.cntRead += stats->tcp.read.cntRead;
+		    current->sock_callstats.read.cntRead += stats->sock_callstats.read.cntRead;
 		    for (ix = 0; ix < 8; ix++) {
-			current->tcp.read.bins[ix] += stats->tcp.read.bins[ix];
+			current->sock_callstats.read.bins[ix] += stats->sock_callstats.read.bins[ix];
 		    }
-		} else if (stats->mTCP == kMode_Client) {
-		    current->tcp.write.WriteErr += stats->tcp.write.WriteErr;
-		    current->tcp.write.WriteCnt += stats->tcp.write.WriteCnt;
+		} else {
+		    current->sock_callstats.write.WriteErr += stats->sock_callstats.write.WriteErr;
+		    current->sock_callstats.write.WriteCnt += stats->sock_callstats.write.WriteCnt;
 #ifdef HAVE_STRUCT_TCP_INFO_TCPI_TOTAL_RETRANS
-		    current->tcp.write.TCPretry += stats->tcp.write.TCPretry;
+		    if (stats->mTCP == kMode_Client) {
+			current->sock_callstats.write.TCPretry += stats->sock_callstats.write.TCPretry;
+		    }
 #endif
 		}
                 if ( current->endTime < stats->endTime ) {
@@ -1002,23 +1303,24 @@ void reporter_handle_multiple_reports( MultiHeader *reporthdr, Transfer_Info *st
 
 #ifdef HAVE_STRUCT_TCP_INFO_TCPI_TOTAL_RETRANS
 static void gettcpistats (ReporterData *stats) {
-    struct tcp_info tcp_internal;
-    socklen_t tcp_info_length = sizeof(struct tcp_info);
-    int retry;
     if (stats->info.mEnhanced && stats->info.mTCP == kMode_Client) {
+	struct tcp_info tcp_internal;
+	socklen_t tcp_info_length = sizeof(struct tcp_info);
+	int retry;
+
 	// Read the TCP retry stats for a client.  Do this
 	// on  a report interval period.
 	if (getsockopt(stats->info.socket, IPPROTO_TCP, TCP_INFO, &tcp_internal, &tcp_info_length) < 0) {
 	    WARN_errno( 1 , "getsockopt");
 	    retry = 0;
 	} else {
-	    retry = tcp_internal.tcpi_total_retrans - stats->info.tcp.write.lastTCPretry;
+	    retry = tcp_internal.tcpi_total_retrans - stats->info.sock_callstats.write.lastTCPretry;
 	}
-	stats->info.tcp.write.TCPretry = retry;
-	stats->info.tcp.write.totTCPretry += retry;
-	stats->info.tcp.write.lastTCPretry = tcp_internal.tcpi_total_retrans;
-	stats->info.tcp.write.cwnd = tcp_internal.tcpi_snd_cwnd * tcp_internal.tcpi_snd_mss / 1024;
-	stats->info.tcp.write.rtt = tcp_internal.tcpi_rtt;
+	stats->info.sock_callstats.write.TCPretry = retry;
+	stats->info.sock_callstats.write.totTCPretry += retry;
+	stats->info.sock_callstats.write.lastTCPretry = tcp_internal.tcpi_total_retrans;
+	stats->info.sock_callstats.write.cwnd = tcp_internal.tcpi_snd_cwnd * tcp_internal.tcpi_snd_mss / 1024;
+	stats->info.sock_callstats.write.rtt = tcp_internal.tcpi_rtt;
     }
 }
 #endif
@@ -1043,23 +1345,37 @@ int reporter_condprintstats( ReporterData *stats, MultiHeader *multireport, int
         stats->info.TotalLen = stats->TotalLen;
         stats->info.startTime = 0;
         stats->info.endTime = TimeDifference( stats->packetTime, stats->startTime );
-	stats->info.transit.minTransit = stats->info.transit.totminTransit;
-	stats->info.transit.maxTransit = stats->info.transit.totmaxTransit;
-	stats->info.transit.cntTransit = stats->info.transit.totcntTransit;
-	stats->info.transit.sumTransit = stats->info.transit.totsumTransit;
-	stats->info.transit.meanTransit = stats->info.transit.totmeanTransit;
-	stats->info.transit.m2Transit = stats->info.transit.totm2Transit;
-	stats->info.transit.vdTransit = stats->info.transit.totvdTransit;
-	if (stats->info.mTCP == kMode_Client) {
-	    stats->info.tcp.write.WriteErr = stats->info.tcp.write.totWriteErr;
-	    stats->info.tcp.write.WriteCnt = stats->info.tcp.write.totWriteCnt;
-	    stats->info.tcp.write.TCPretry = stats->info.tcp.write.totTCPretry;
+	// There is a corner case when the first packet is also the last where the start time (which comes
+	// from app level syscall) is greater than the packetTime (which come for kernel level SO_TIMESTAMP)
+	// For this case set the start and end time to both zero.
+	if (stats->info.endTime < 0) {
+	    stats->info.endTime = 0;
+	}
+	if (stats->info.mUDP == kMode_Server) {
+	    stats->info.l2counts.cnt = stats->info.l2counts.tot_cnt;
+	    stats->info.l2counts.unknown = stats->info.l2counts.tot_unknown;
+	    stats->info.l2counts.udpcsumerr = stats->info.l2counts.tot_udpcsumerr;
+	    stats->info.l2counts.lengtherr = stats->info.l2counts.tot_lengtherr;
+	    stats->info.transit.minTransit = stats->info.transit.totminTransit;
+	    stats->info.transit.maxTransit = stats->info.transit.totmaxTransit;
+	    stats->info.transit.cntTransit = stats->info.transit.totcntTransit;
+	    stats->info.transit.sumTransit = stats->info.transit.totsumTransit;
+	    stats->info.transit.meanTransit = stats->info.transit.totmeanTransit;
+	    stats->info.transit.m2Transit = stats->info.transit.totm2Transit;
+	    stats->info.transit.vdTransit = stats->info.transit.totvdTransit;
+	}
+	if ((stats->info.mTCP == kMode_Client) || (stats->info.mUDP == kMode_Client)) {
+	    stats->info.sock_callstats.write.WriteErr = stats->info.sock_callstats.write.totWriteErr;
+	    stats->info.sock_callstats.write.WriteCnt = stats->info.sock_callstats.write.totWriteCnt;
+	    if (stats->info.mTCP == kMode_Client) {
+		stats->info.sock_callstats.write.TCPretry = stats->info.sock_callstats.write.totTCPretry;
+	    }
 	}
 	if (stats->info.mTCP == kMode_Server) {
 	    int ix;
-	    stats->info.tcp.read.cntRead = stats->info.tcp.read.totcntRead;
+	    stats->info.sock_callstats.read.cntRead = stats->info.sock_callstats.read.totcntRead;
 	    for (ix = 0; ix < 8; ix++) {
-		stats->info.tcp.read.bins[ix] = stats->info.tcp.read.totbins[ix];
+		stats->info.sock_callstats.read.bins[ix] = stats->info.sock_callstats.read.totbins[ix];
 	    }
 	}
 	if (stats->info.endTime > 0) {
@@ -1069,7 +1385,13 @@ int reporter_condprintstats( ReporterData *stats, MultiHeader *multireport, int
 	}
 	stats->info.IPGsum = 1;
         stats->info.free = 1;
-
+#ifdef HAVE_ISOCHRONOUS
+	if (stats->info.mIsochronous) {
+	    stats->info.isochstats.framecnt = stats->isochstats.framecnt;
+	    stats->info.isochstats.framelostcnt = stats->isochstats.framelostcnt;
+	    stats->info.isochstats.slipcnt = stats->isochstats.slipcnt;
+	}
+#endif
         reporter_print( stats, TRANSFER_REPORT, force );
         if ( isMultipleReport(stats) ) {
             reporter_handle_multiple_reports( multireport, &stats->info, force );
@@ -1111,16 +1433,22 @@ int reporter_condprintstats( ReporterData *stats, MultiHeader *multireport, int
 	    if (stats->info.mUDP) {
 		stats->info.IPGcnt = 0;
 		stats->info.IPGsum = 0;
+		if (stats->info.mUDP == kMode_Server) {
+		    stats->info.l2counts.cnt = 0;
+		    stats->info.l2counts.unknown = 0;
+		    stats->info.l2counts.udpcsumerr = 0;
+		    stats->info.l2counts.lengtherr = 0;
+		}
 	    }
 	    if (stats->info.mEnhanced) {
-		if (stats->info.mTCP == (char)kMode_Client) {
-		    stats->info.tcp.write.WriteCnt = 0;
-		    stats->info.tcp.write.WriteErr = 0;
+		if ((stats->info.mTCP == (char)kMode_Client) || (stats->info.mUDP == (char)kMode_Client)) {
+		    stats->info.sock_callstats.write.WriteCnt = 0;
+		    stats->info.sock_callstats.write.WriteErr = 0;
 		} else if (stats->info.mTCP == (char)kMode_Server) {
 		    int ix;
-		    stats->info.tcp.read.cntRead = 0;
+		    stats->info.sock_callstats.read.cntRead = 0;
 		    for (ix = 0; ix < 8; ix++) {
-			stats->info.tcp.read.bins[ix] = 0;
+			stats->info.sock_callstats.read.bins[ix] = 0;
 		    }
 		}
 	    }
