@@ -66,14 +66,31 @@
 #include "Settings.hpp"
 #include "Locale.h"
 #include "SocketAddr.h"
-
 #include "util.h"
 #include "version.h"
-
 #include "gnu_getopt.h"
+#ifdef HAVE_ISOCHRONOUS
+#include "isochronous.hpp"
+#include "pdfs.h"
+#endif
+#ifdef HAVE_UDPTRIGGERS
+#include "ioctls.h"
+#endif
 
 static int seqno64b = 0;
 static int reversetest = 0;
+static int udphistogram = 0;
+static int l2checks = 0;
+static int incrdstip = 0;
+static int txsync = 0;
+#ifdef HAVE_UDPTRIGGERS
+static int udptriggers = 0;
+#endif
+#ifdef HAVE_ISOCHRONOUS
+static int burstipg = 0;
+static int burstipg_set = 0;
+static int isochronous = 0;
+#endif
 
 void Settings_Interpret( char option, const char *optarg, thread_Settings *mExtSettings );
 // apply compound settings after the command line has been fully parsed
@@ -118,6 +135,7 @@ const struct option long_options[] =
 {"compatibility",    no_argument, NULL, 'C'},
 {"daemon",           no_argument, NULL, 'D'},
 {"file_input", required_argument, NULL, 'F'},
+{"ssm-host", required_argument, NULL, 'H'},
 {"stdin_input",      no_argument, NULL, 'I'},
 {"mss",        required_argument, NULL, 'M'},
 {"nodelay",          no_argument, NULL, 'N'},
@@ -136,6 +154,17 @@ const struct option long_options[] =
 {"peer-detect",      no_argument, NULL, 'X'},
 {"linux-congestion", required_argument, NULL, 'Z'},
 {"udp-counters-64bit", no_argument, &seqno64b, 1},
+{"udp-histogram", optional_argument, &udphistogram, 1},
+{"l2checks", no_argument, &l2checks, 1},
+{"incr-dstip", no_argument, &incrdstip, 1},
+{"tx-sync", required_argument, &txsync, 1},
+#ifdef HAVE_UDPTRIGGERS
+{"udp-triggers", no_argument, &udptriggers, 1},
+#endif
+#ifdef HAVE_ISOCHRONOUS
+{"ipg", required_argument, &burstipg, 1},
+{"isochronous", optional_argument, &isochronous, 1},
+#endif
 #ifdef WIN32
 {"reverse", no_argument, &reversetest, 1},
 #endif
@@ -189,7 +218,7 @@ const struct option env_options[] =
 
 #define SHORT_OPTIONS()
 
-const char short_options[] = "1b:c:def:hi:l:mn:o:p:rst:uvw:x:y:zB:CDF:IL:M:NP:RS:T:UVWXZ:";
+const char short_options[] = "1b:c:def:hi:l:mn:o:p:rst:uvw:x:y:zB:CDF:H:IL:M:NP:RS:T:UVWXZ:";
 
 /* -------------------------------------------------------------------
  * defaults
@@ -215,6 +244,7 @@ void Settings_Initialize( thread_Settings *main ) {
     main->mReportMode = kReport_Default;
     // option, defaults
     main->flags         = FLAG_MODETIME | FLAG_STDOUT; // Default time and stdout
+    main->flags_extend  = 0x0;           // Default all extend flags to off
     //main->mUDPRate      = 0;           // -b,  offered (or rate limited) load (both UDP and TCP)
     main->mUDPRateUnits = kRate_BW;
     //main->mHost         = NULL;        // -c,  none, required for client
@@ -249,7 +279,7 @@ void Settings_Initialize( thread_Settings *main ) {
     //main->mThreads      = 0;           // -P,
     //main->mRemoveService = false;      // -R,
     //main->mTOS          = 0;           // -S,  ie. don't set type of service
-    main->mTTL          = 1;             // -T,  link-local TTL
+    main->mTTL          = -1;            // -T,  link-local TTL
     //main->mDomain     = kMode_IPv4;    // -V,
     //main->mSuggestWin = false;         // -W,  Suggest the window size.
 
@@ -274,10 +304,31 @@ void Settings_Copy( thread_Settings *from, thread_Settings **into ) {
         (*into)->mFileName = new char[ strlen(from->mFileName) + 1];
         strcpy( (*into)->mFileName, from->mFileName );
     }
+    if ( from->mUDPHistogramStr != NULL ) {
+	(*into)->mUDPHistogramStr = new char[ strlen(from->mUDPHistogramStr) + 1];
+        strcpy( (*into)->mUDPHistogramStr, from->mUDPHistogramStr );
+    }
+    if ( from->mSSMMulticastStr != NULL ) {
+	(*into)->mSSMMulticastStr = new char[ strlen(from->mSSMMulticastStr) + 1];
+        strcpy( (*into)->mSSMMulticastStr, from->mSSMMulticastStr );
+    }
+    if ( from->mIfrname != NULL ) {
+	(*into)->mIfrname = new char[ strlen(from->mIfrname) + 1];
+        strcpy( (*into)->mIfrname, from->mIfrname );
+    }
+#ifdef HAVE_ISOCHRONOUS
+    if ( from->mIsochronousStr != NULL ) {
+	(*into)->mIsochronousStr = new char[ strlen(from->mIsochronousStr) + 1];
+        strcpy( (*into)->mIsochronousStr, from->mIsochronousStr );
+    }
+#endif
     // Zero out certain entries
     (*into)->mTID = thread_zeroid();
     (*into)->runNext = NULL;
     (*into)->runNow = NULL;
+#if defined(HAVE_LINUX_FILTER_H) && defined(HAVE_AF_PACKET)
+    (*into)->mSockDrop = INVALID_SOCKET;
+#endif
 }
 
 /* -------------------------------------------------------------------
@@ -289,6 +340,12 @@ void Settings_Destroy( thread_Settings *mSettings) {
     DELETE_ARRAY( mSettings->mLocalhost );
     DELETE_ARRAY( mSettings->mFileName  );
     DELETE_ARRAY( mSettings->mOutputFileName );
+    DELETE_ARRAY( mSettings->mUDPHistogramStr );
+    DELETE_ARRAY( mSettings->mSSMMulticastStr);
+    DELETE_ARRAY( mSettings->mIfrname);
+#ifdef HAVE_ISOCHRONOUS
+    DELETE_ARRAY( mSettings->mIsochronousStr );
+#endif
     DELETE_PTR( mSettings );
 } // end ~Settings
 
@@ -335,37 +392,33 @@ void Settings_ParseCommandLine( int argc, char **argv, thread_Settings *mSetting
  * ------------------------------------------------------------------- */
 
 void Settings_Interpret( char option, const char *optarg, thread_Settings *mExtSettings ) {
-    max_size_t theNum;
-
+    char *results;
     switch ( option ) {
         case '1': // Single Client
             setSingleClient( mExtSettings );
             break;
+
         case 'b': // UDP bandwidth
-	    // scan for PPS units, just look for 'p' as that's good enough
 	    {
-		char *end;
-		errno = 0;
-#ifdef HAVE_STRTOLL
-		theNum = strtoll (optarg, &end, 10);
-#else
-		theNum = strtol (optarg, &end, 10);
-#endif
-		if (errno || end == optarg) {
-		    fprintf(stderr, "Invalid bandwidth -b of %s\n", optarg);
-		    exit(1);
-		}
-		if (*end == 'p' || *end == 'P') {
+		char *tmp= new char [strlen(optarg) + 1];
+		strcpy(tmp, optarg);
+		// scan for PPS units, just look for 'p' as that's good enough
+		if ((((results = strtok(tmp, "p")) != NULL) && strcmp(results,optarg)) \
+		    || (((results = strtok(tmp, "P")) != NULL)  && strcmp(results,optarg))) {
 		    mExtSettings->mUDPRateUnits = kRate_PPS;
-		    mExtSettings->mUDPRate = theNum;
+		    mExtSettings->mUDPRate = byte_atoi(results);
 		} else {
 		    mExtSettings->mUDPRateUnits = kRate_BW;
 		    mExtSettings->mUDPRate = byte_atoi(optarg);
+		    if (((results = strtok(tmp, ",")) != NULL) && strcmp(results,optarg)) {
+			setVaryLoad(mExtSettings);
+			mExtSettings->mVariance = byte_atoi(optarg);
+		    }
 		}
+		delete [] tmp;
 	    }
-            setBWSet( mExtSettings );
-            break;
-
+	    setBWSet( mExtSettings );
+	    break;
         case 'c': // client mode w/ server host to connect to
             mExtSettings->mHost = new char[ strlen( optarg ) + 1 ];
             strcpy( mExtSettings->mHost, optarg );
@@ -570,6 +623,16 @@ void Settings_Interpret( char option, const char *optarg, thread_Settings *mExtS
             strcpy( mExtSettings->mFileName, optarg);
             break;
 
+        case 'H' : // Get the SSM host (or Source per the S,G)
+            if ( mExtSettings->mThreadMode == kMode_Client ) {
+                fprintf( stderr, warn_invalid_client_option, option );
+                break;
+            }
+            mExtSettings->mSSMMulticastStr = new char[strlen(optarg)+1];
+            strcpy( mExtSettings->mSSMMulticastStr, optarg);
+            setSSMMulticast( mExtSettings );
+            break;
+
         case 'I' : // Set the stdin as the input source
             if ( mExtSettings->mThreadMode != kMode_Client ) {
                 fprintf( stderr, warn_invalid_server_option, option );
@@ -629,7 +692,7 @@ void Settings_Interpret( char option, const char *optarg, thread_Settings *mExtS
             mExtSettings->mTOS = strtol( optarg, NULL, 0 );
             break;
 
-        case 'T': // time-to-live for multicast
+        case 'T': // time-to-live for both unicast and multicast
             mExtSettings->mTTL = atoi( optarg );
             break;
 
@@ -667,17 +730,82 @@ void Settings_Interpret( char option, const char *optarg, thread_Settings *mExtS
 
         case 0:
 	    if (seqno64b) {
-#ifdef HAVE_SEQNO64b
+		seqno64b = 0;
+#if (HAVE_SEQNO64b && (HAVE_QUAD_SUPPORT || HAVE_INT64_T))
 		setSeqNo64b(mExtSettings);
 #else
 		fprintf( stderr, "WARNING: 64 bit sequence numbers not supported\n");
 #endif
 	    }
+	    if (incrdstip) {
+		incrdstip = 0;
+		setIncrDstIP(mExtSettings);
+	    }
+	    if (txsync) {
+		txsync = 0;
+		setTxSync(mExtSettings);
+		char *end;
+		mExtSettings->mTxSyncInterval = strtof(optarg,&end);
+		if (*end != '\0') {
+		    fprintf (stderr, "Invalid value of '%s' for --tx-sync\n", optarg);
+		}
+	    }
+	    if (udphistogram) {
+		udphistogram = 0;
+		setUDPHistogram( mExtSettings );
+		setEnhanced( mExtSettings );
+		// The following are default values which
+		mExtSettings->mUDPbins = 1000;
+		mExtSettings->mUDPbinsize = 1;
+		mExtSettings->mUDPunits = 0;
+		mExtSettings->mUDPci_lower = 5;
+		mExtSettings->mUDPci_upper = 95;
+		if (optarg) {
+		    mExtSettings->mUDPHistogramStr = new char[ strlen( optarg ) + 1 ];
+		    strcpy(mExtSettings->mUDPHistogramStr, optarg);
+		}
+	    }
 	    if (reversetest) {
+		reversetest = 0;
 		fprintf( stderr, "WARNING: The --reverse option is currently not supported\n");
 		exit(1);
 		setReverse(mExtSettings);
 	    }
+#ifdef HAVE_ISOCHRONOUS
+	    if (isochronous) {
+		isochronous = 0;
+		setEnhanced( mExtSettings );
+		setIsochronous(mExtSettings);
+		// The following are default values which
+		// may be overwritten during modal parsing
+		mExtSettings->mFPS = 60.0;
+		mExtSettings->mMean = 20000000.0;
+		mExtSettings->mVariance = 0.0;
+		mExtSettings->mBurstIPG = 0.005;
+		if (optarg) {
+		    mExtSettings->mIsochronousStr = new char[ strlen( optarg ) + 1 ];
+		    strcpy( mExtSettings->mIsochronousStr, optarg );
+		}
+	    }
+	    if (burstipg) {
+		burstipg = 0;
+		burstipg_set = 1;
+		char *end;
+		mExtSettings->mBurstIPG = strtof(optarg,&end);
+		if (*end != '\0') {
+		    fprintf (stderr, "Invalid value of '%s' for --ipg\n", optarg);
+		}
+	    }
+#endif
+#ifdef HAVE_UDPTRIGGERS
+	    if (udptriggers) {
+		udptriggers = 0;
+		setUDP( mExtSettings );
+		setUDPTriggers(mExtSettings);
+		setSeqNo64b(mExtSettings);  // enable this if udp triggers
+	    }
+#endif
+	    break;
         default: // ignore unknown
             break;
     }
@@ -698,6 +826,7 @@ void Settings_Interpret( char option, const char *optarg, thread_Settings *mExtS
 //  Other things that need this are multicast socket or not,
 //  -B local bind port parsing, and when to use the default UDP offered load
 void Settings_ModalOptions( thread_Settings *mExtSettings ) {
+    char *results;
     // Handle default read/write sizes based on v4, v6, UDP or TCP
     if ( !isBuflenSet( mExtSettings ) ) {
 	if (isUDP(mExtSettings)) {
@@ -714,23 +843,127 @@ void Settings_ModalOptions( thread_Settings *mExtSettings ) {
     if (!isBWSet(mExtSettings) && isUDP(mExtSettings)) {
 	mExtSettings->mUDPRate = kDefault_UDPRate;
     }
-    // Check for local port assignment via parsing -B's mLocalhost string
-    // (only supported on the client as server/listener uses -p for this)
-    if ( mExtSettings->mLocalhost != NULL && mExtSettings->mThreadMode == kMode_Client ) {
-	char *results;
-	// v4 uses a colon as the delimeter for the local bind port, e.g. 192.168.1.1:6001
-	if (!isIPV6(mExtSettings)) {
-	    if (((results = strtok(mExtSettings->mLocalhost, ":")) != NULL) && ((results = strtok(NULL, ":")) != NULL)) {
-		mExtSettings->mBindPort = atoi(results);
+    // Check options for mutualities
+    if (isTxSync(mExtSettings)) {
+	if (mExtSettings->mThreadMode != kMode_Client) {
+	    fprintf(stderr, "option --tx-sync only supported on clients\n");
+	    exit(1);
+	}
+	if (isUDP(mExtSettings)) {
+	    if (isBWSet(mExtSettings)) {
+		fprintf(stderr, "option --tx-sync and -b are mutually exclusive\n");
+		exit(1);
 	    }
-	// v6 uses bracket notation, e.g. [2001:e30:1401:2:d46e:b891:3082:b939]:6001
-	} else if (mExtSettings->mLocalhost[0] ==  '[') {
-	    if ((results = strtok(mExtSettings->mLocalhost, "]")) != NULL) {
-		results++;
-		strcpy(mExtSettings->mLocalhost, results);
-		if ((results = strtok(NULL, ":")) != NULL) {
+	} else {
+	    fprintf(stderr, "option --tx-sync not supported for TCP, only UDP (-u option)\n");
+	    exit(1);
+	}
+	if (isIsochronous(mExtSettings)) {
+	  fprintf(stderr, "option --tx-sync and --isochronous are mutually exclusive\n");
+	  exit(1);
+	}
+    }
+    if ((mExtSettings->mThreadMode != kMode_Client) && isVaryLoad(mExtSettings)) {
+	fprintf(stderr, "option of variance ignored as not supported on the server\n");
+    }
+
+    // UDP histogram settings
+    if (isUDPHistogram(mExtSettings) && isUDP(mExtSettings) && mExtSettings->mThreadMode != kMode_Client) {
+	if (((results = strtok(mExtSettings->mUDPHistogramStr, ",")) != NULL) && !strcmp(results,mExtSettings->mUDPHistogramStr)) {
+	    char *tmp = new char [strlen(results) + 1];
+	    strcpy(tmp, results);
+	    // scan for microseconds as units
+	    if ((strtok(tmp, "u") != NULL) && strcmp(results,tmp)) {
+		mExtSettings->mUDPunits = 1;
+	    }
+	    mExtSettings->mUDPbinsize = atoi(tmp);
+	    delete [] tmp;
+	    if ((results = strtok(results+strlen(results)+1, ",")) != NULL) {
+		mExtSettings->mUDPbins = byte_atoi(results);
+		if ((results = strtok(NULL, ",")) != NULL) {
+		    mExtSettings->mUDPci_lower = atoi(results);
+		    if ((results = strtok(NULL, ",")) != NULL) {
+			mExtSettings->mUDPci_upper = atoi(results);
+		    }
+		}
+	    }
+	}
+    }
+    // L2 settings
+    if (l2checks && isUDP(mExtSettings)) {
+	l2checks = 0;
+
+	// Client controls hash or not
+	if (mExtSettings->mThreadMode == kMode_Client) {
+	    setL2LengthCheck(mExtSettings);
+	} else {
+#if defined(HAVE_LINUX_FILTER_H) && defined(HAVE_AF_PACKET)
+	  // Request server to do length checks
+	  setL2LengthCheck(mExtSettings);
+#else
+	  fprintf(stderr, "--l2checks not supported on this platform\n");
+#endif
+	}
+    }
+
+
+#ifdef HAVE_ISOCHRONOUS
+    if (mExtSettings->mBurstIPG > 0.0) {
+	if (!isIsochronous(mExtSettings)) {
+	    fprintf(stderr, "option --ipg requires the --isochronous option\n");
+	    exit(1);
+	}
+	if (mExtSettings->mThreadMode != kMode_Client) {
+	    fprintf(stderr, "option --ipg only supported on clients\n");
+	    exit(1);
+	}
+    }
+    if (isIsochronous(mExtSettings) && mExtSettings->mIsochronousStr) {
+	// parse client isochronous field,
+	// format is --isochronous <int>:<float>,<float> and supports
+	// human suffixes, e.g. --isochronous 60:100m,5m
+	// which is frames per second, mean and variance
+	if (mExtSettings->mThreadMode == kMode_Client) {
+	    if (((results = strtok(mExtSettings->mIsochronousStr, ":")) != NULL) && !strcmp(results,mExtSettings->mIsochronousStr)) {
+		mExtSettings->mFPS = atof(results);
+		if ((results = strtok(NULL, ",")) != NULL) {
+		    mExtSettings->mMean = bitorbyte_atof(results);
+		    if ((results = strtok(NULL, ",")) != NULL) {
+			mExtSettings->mVariance = bitorbyte_atof(results);
+		    }
+		} else {
+		    mExtSettings->mMean = 20000000.0;
+		    mExtSettings->mVariance = 0.0;
+		}
+	    } else {
+		fprintf(stderr, "Invalid --isochronous value, format is <fps>:<mean>,<variance> (e.g. 60:18M,1m)\n");
+	    }
+	}
+    }
+#endif
+    // Check for further mLocalhost (-B) parsing:
+    if ( mExtSettings->mLocalhost) {
+	// Check for -B device
+	if (((results = strtok(mExtSettings->mLocalhost, "%")) != NULL) && ((results = strtok(NULL, "%")) != NULL)) {
+	    mExtSettings->mIfrname = new char[ strlen(results) + 1 ];
+	    strcpy( mExtSettings->mIfrname, results );
+	}
+	// Client local host parsing
+	if (mExtSettings->mThreadMode == kMode_Client ) {
+	    // v4 uses a colon as the delimeter for the local bind port, e.g. 192.168.1.1:6001
+	    if (!isIPV6(mExtSettings)) {
+		if (((results = strtok(mExtSettings->mLocalhost, ":")) != NULL) && ((results = strtok(NULL, ":")) != NULL)) {
 		    mExtSettings->mBindPort = atoi(results);
 		}
+		// v6 uses bracket notation, e.g. [2001:e30:1401:2:d46e:b891:3082:b939]:6001
+	    } else if (mExtSettings->mLocalhost[0] ==  '[') {
+		if ((results = strtok(mExtSettings->mLocalhost, "]")) != NULL) {
+		    results++;
+		    strcpy(mExtSettings->mLocalhost, results);
+		    if ((results = strtok(NULL, ":")) != NULL) {
+			mExtSettings->mBindPort = atoi(results);
+		    }
+		}
 	    }
 	}
     }
@@ -817,13 +1050,11 @@ void Settings_GenerateListenerSettings( thread_Settings *client, thread_Settings
 }
 
 /*
- * Settings_GenerateSpeakerSettings
- * Called to generate the settings to be passed to the Speaker
- * instance that will handle dual testings from the server side
- * this should only return an instance if it was called on
- * the thread_Settings instance generated from the command line
- * for server side execution. This should be an inverse operation
- * of GenerateClientHdr.
+ * Settings_GenerateClientSettings
+ *
+ * Called by the Listener to generate the settings to be used by clients
+ * per things like dual tests.
+ *
  */
 void Settings_GenerateClientSettings( thread_Settings *server,
                                       thread_Settings **client,
@@ -894,14 +1125,17 @@ void Settings_GenerateClientSettings( thread_Settings *server,
 
 /*
  * Settings_GenerateClientHdr
- * Called to generate the client header to be passed to the
- * server that will handle dual testings from the server side
- * This should be an inverse operation of GenerateSpeakerSettings
+ *
+ * Called to generate the client header to be passed to the listener/server
+ *
+ * This will handle:
+ * o) dual testings from the listener/server side
+ * o) advanced udp test settings
  *
  * Returns hdr flags set
  */
 int Settings_GenerateClientHdr( thread_Settings *client, client_hdr *hdr ) {
-    int flags = 0, extendflags = 0;
+    uint32_t flags = 0, extendflags = 0;
     if (isPeerVerDetect(client) || (client->mMode != kTest_Normal && isBWSet(client))) {
 	flags |= HEADER_EXTEND;
     }
@@ -928,6 +1162,46 @@ int Settings_GenerateClientHdr( thread_Settings *client, client_hdr *hdr ) {
 	    flags |= RUN_NOW;
 	}
     }
+    if (isUDP(client)) {
+	/*
+	 * set the default offset where underlying "inline" subsystems can write into the udp payload
+	 */
+	hdr->udp.tlvoffset = htons((sizeof(client_hdr_udp_tests) + sizeof(client_hdr_v1) + sizeof(UDP_datagram)));
+
+	if (isL2LengthCheck(client) || isIsochronous(client) || isUDPTriggers(client)) {
+	    flags |= HEADER_UDPTESTS;
+	    uint16_t testflags = 0;
+
+	    if (isL2LengthCheck(client)) {
+		testflags |= HEADER_L2LENCHECK;
+		if (isIPV6(client))
+		    testflags |= HEADER_L2ETHPIPV6;
+	    }
+	    if (isIsochronous(client)) {
+		hdr->udp.tlvoffset = htons((sizeof(UDP_isoch_payload) + sizeof(client_hdr_udp_tests) + sizeof(client_hdr_v1) + sizeof(UDP_datagram)));
+		testflags |= HEADER_UDP_ISOCH;
+	    }
+	    if (isUDPTriggers(client)) {
+		testflags |= HEADER_UDPTRIGGERS;
+		hdr->udp.ref_sync_tv_sec = htonl(0xFFFFFFFF);
+	        hdr->udp.ref_sync_tv_nsec = htonl(0x0);
+#ifdef HAVE_UDPTRIGGERS
+		if (client->mIfrname) {
+		    uint32_t tsfnow = read_80211_tsf(client);
+		    hdr->udp.ref_sync_tv_sec = htonl(tsfnow/1000000);
+		    hdr->udp.ref_sync_tv_nsec = htonl((tsfnow%1000000) * 1000);
+		    Timestamp gpsnow;
+		    hdr->udp.gps_sync_tv_sec = htonl(gpsnow.getSecs());
+		    hdr->udp.gps_sync_tv_nsec = htonl(gpsnow.getUsecs() * 1000);
+		}
+#endif
+	    }
+	    // Write flags to header so the listener can determine the tests requested
+	    hdr->udp.testflags = htons(testflags);
+	    hdr->udp.version_u = htonl(IPERF_VERSION_MAJORHEX);
+	    hdr->udp.version_l = htonl(IPERF_VERSION_MINORHEX);
+	}
+    }
     /*
      * Finally, update the header flags (to be passed to the remote server)
      */
